"""
Implements a data pump for extracted file data which allows for
more robust file identification, reporting, and objectifying
content to ease maintenance.
"""

import logging
from collections import deque

from mwcp.file_object import FileObject
from mwcp.parser import Parser

logger = logging.getLogger(__name__)


class UnableToParse(Exception):
    """
    This exception can be thrown if a parser that has been correctly identified has failed to parse
    the file and you would like other parsers to be tried.
    """

    pass


class UnidentifiedFile(Parser):
    """Describes an unidentified file. This parser will hit on any FileObject."""

    DESCRIPTION = "Unidentified file"

    @classmethod
    def identify(cls, file_object):
        """
        Identifies an unidentified file... which means this is always True.

        :param file_object: dispatcher.FileObject object
        :return: Boolean indicating idenification
        """
        return True


# NOTE: This object is duck typed to look like a Parser object to allow recursion.
# TODO: Create abstract inteface for these two.
class Dispatcher(object):
    # TODO: Rewrite this documentation.
    """
    This class will continuously process items that are in the queue.  When the queue is empty,
    this will ultimately signal that processing is complete and the script will terminate.
    This class will process the items using the supplied list of Parser classes provided.

    This class can be used as a mixin along with the Parser class or
    can be initialized by itself.
    When used as a mixin, the dispatcher will automatically add the file in the reporter
    to the queue and run dispatch() when run() is called.

    """

    def __init__(
        self,
        name,
        source,
        author="",
        description="",
        parsers=None,
        greedy=False,
        default=UnidentifiedFile,
        output_unidentified=True,
        overwrite_descriptions=False,
        embedded=False,
    ):
        """
        Initializes the Dispatcher with the given parsers to run.

        :param str name: Unique name to give to this group of parsers.
        :param str source: Name of source this group comes from.
        :param str author: Author of the parser.
        :param str description: Description of the parser.
        :param list parsers: A list of parser classes (or other dispatchers) to use for detection and running.
            Order of this list is the order the Dispatcher will perform its identification.
            If not provided, it will default to an empty list.
        :param bool greedy: By default, the dispatcher will only run on the first parser it detects
            to be a valid parser. If greedy is set to true, the dispatcher will try all parsers
            even if a previous parser was successful.
        :param ComponentParser default: The Parser class to default to if no parsers in the parsers list
            has identified it. If set to None, no parser will be run as default.
            (By default, the dispatcher.UnidentifiedFile will be run.)
        :param bool output_unidentified: Whether to output files that have not been identified by
            any parsers.
        :param bool overwrite_descriptions: Whether to allow dispatcher to overwrite any previous
            set description with the parser's
        :param bool embedded: If True, all dispatched files will be passed up to the parent
            dispatcher instead of being processed locally.
            Ie, this is the equivalent of embedding the listed parsers directly into the parent's list.
        """
        self.name = name
        self.source = source
        # TODO: Deprecate the author attribute?
        self.AUTHOR = author
        self.DESCRIPTION = description  # In all caps to match Parser interface
        self.parsers = parsers or []
        self.greedy = greedy
        self.default = default
        self._fifo_buffer = deque()
        self._current_file_object = None
        self._current_parser = None
        self._output_unidentified = output_unidentified
        self._overwrite_descriptions = overwrite_descriptions
        self._embedded = embedded

        # Dictionary that can be used by parsers to pass variables across parsers.
        # E.g. an encryption key found in the loader to be used by the implant.
        self.knowledge_base = {}

    def __repr__(self):
        return "{}({})".format(self.name, ", ".join(repr(parser) for parser in self.parsers))

    def identify(self, file_object):
        """
        Determines if this dispatcher is identified to support the given file_object.

        :param file_object: file object to use for identification
        :type file_object: dispatcher.FileObject

        :return bool: Boolean indicating if this dispatcher supports the file_object
        """
        return any(parser.identify(file_object) for parser in self.parsers)

    def add_to_queue(self, file_object: FileObject, parent: FileObject = None):
        """
        Add a FileObject to the FIFO queue for processing.
        :param file_object: a FileObject object requiring processing.
        :param parent: original parent for given file_object.
            If not provided, the parent is assumed to be the file being currently
            processed. (which should be true most of the time)

        :return:
        """
        if not parent:
            parent = self._current_file_object
        assert isinstance(file_object, FileObject), "Not a FileObject: {!r}".format(file_object)
        # If we already have a parent, this means this file is trickling up from a sub-dispatcher.
        # Don't duplicate logs or change the parent.
        if not file_object.parent:
            file_object.parent = parent
            if parent:
                parent.children.append(file_object)
                logger.info(f"{parent.name} dispatched residual file: {file_object.name}")
                if file_object.description:
                    logger.info(f"File {file_object.name} described as {file_object.description}")

        self._fifo_buffer.appendleft(file_object)

    def _iter_parsers(self, file_object):
        """
        Generator that detects and yields applicable parsers to run based on given file_object.

        :param FileObject file_object: file object that needs to be identified

        :yields: Identified Parser class or another Dispatcher that can be run
        """
        for parser in self.parsers:
            logger.debug(u"Identifying {} with {!r}.".format(file_object.name, parser))
            if parser.identify(file_object):
                if isinstance(parser, Dispatcher):
                    # Parser is a group, change wording
                    logger.info(f"File {file_object.name} identified with {parser.DESCRIPTION} parser.")
                else:
                    logger.info(f"File {file_object.name} identified as {parser.DESCRIPTION}.")
                logger.debug(u"{} identified with {!r}".format(file_object.name, parser))
                yield parser

    def _parse(self, file_object, parser, reporter):
        """
        Parse given file_object with given sub parser

        :raises UnableToParse: If the subparser raised an error.
        """
        self._current_file_object = file_object
        self._current_parser = parser

        # If a description wasn't set for the file, use the parser's
        # (But ignore setting it for sub dispatchers)
        if (not file_object.description or self._overwrite_descriptions) and not isinstance(parser, Dispatcher):
            file_object.description = parser.DESCRIPTION

        # Set parser class used in order to keep a history.
        file_object.parser = parser

        try:
            parser.parse(file_object, reporter, dispatcher=self)
        except UnableToParse as exception:
            if isinstance(parser, Dispatcher):
                # Parser is a group, change wording
                logger.info(
                    f"File {file_object.file_name} was misidentified with {parser.DESCRIPTION} parser, due to: "
                    f"({exception}) Trying other parsers..."
                )
            else:
                logger.info(
                    f"File {file_object.file_name} was misidentified as {parser.DESCRIPTION}, due to: "
                    f"({exception}) Trying other parsers..."
                )
            raise
        except Exception:
            logger.exception(u"{} dispatch parser failed".format(parser.name))

    def parse(self, file_object, reporter, dispatcher=None):
        """
        Runs dispatcher on given file_object.

        :param FileObject file_object: Object containing data about component file.
        :param mwcp.Reporter reporter: reference to reporter object that executed this parser.
        :param Dispatcher dispatcher: reference to the parent dispatcher object that called this parse command.
            (None if this dispatcher is the root)
        :return:
        """
        # TODO: Use reporter to output metadata about initial input file.
        self.add_to_queue(file_object)

        # Pull knowledge_base from previous dispatcher.
        if dispatcher:
            self.knowledge_base = dispatcher.knowledge_base

        first = True
        while self._fifo_buffer:
            file_object = self._fifo_buffer.pop()

            # If this dispatcher is embedded, simply pass any dispatched files to the parent.
            if self._embedded and dispatcher and not first:
                dispatcher.add_to_queue(file_object)
                continue

            first = False
            identified = False

            try:
                unable_to_parse_error = None
                # Run any applicable parsers.
                for parser in self._iter_parsers(file_object):
                    try:
                        self._parse(file_object, parser, reporter)
                    except UnableToParse as e:
                        unable_to_parse_error = e
                        continue
                    identified = True
                    if not self.greedy:
                        break
                if identified:
                    continue
                elif unable_to_parse_error and dispatcher:
                    # Pass UnableToParse exception up the chain to notify parent
                    raise unable_to_parse_error

                # Give it to the parent dispatcher if we can't identify it.
                if dispatcher:
                    dispatcher.add_to_queue(file_object)
                    continue

                # If no parsers match and developer didn't set a description,
                # mark as unidentified file and run default.
                if not file_object.description:
                    logger.info(u"Supplied file {} was not identified.".format(file_object.file_name))
                    if self.default:
                        try:
                            self._parse(file_object, self.default, reporter)
                        except UnableToParse:
                            pass

            finally:
                # Output the file if we identified it or we are the root.
                # NOTE: We don't want to output the file until the very end, since a parser may want to change
                # the file's filename or description.
                if identified or (not dispatcher and self._output_unidentified):
                    file_object.output()



"""
Collection of patches done to bring back some of the removed features of 2.8 back into 2.9
as well as generally fix lingering issues with construct.

To activate, replace your standard import with this:
    from mwcp.utils.construct import version28 as construct

Patches:
    - slicing mechanism ([:], [min:], [:max], etc)
    - allows default value for pop() in Containers
    - allow any encoding for string constructs.
    - patch Embedded to remove hardcoded limitation of supported classes.
    - fixes issue with sizeof used with dynamic Structs (issue #771)
    - patch StringEncoded to make UnicodeDecodeErrors as StringError (issue #743)

Also contains fixes for few constructs. (Use these versions to get the benefits.)
    - Range() - was removed in 2.9
    - PaddedString() - provides default encoding of 'utf-8' and fixes limitation on encoding codecs
    - String() - alias to PaddedString
    - GreedyString - default encoding to 'utf-8'
    - Compressed() - add ability to provide any algorithm (provided it has a compress() and decompress() function)
    - Union() - Allows for the parsefrom parameter to be optional.
    - Probe() - changed to show lookahead stream by default since that is almost always preferred.
              - Also converted back to use hexdump instead of just showing a hex string.
    - FocusedSeq() - Adds back support for supplying an index and fixes building.
    - Mapping() - Patches Mapping to allow non-symmetric mappings again.


Wishlist: (the following are things we would like to fix in version 2.9 but are out of scope for now)
    - Go back to merging embedded fields after parsing instead of before. This change breaks the ability
        to embed any Construct type. (e.g. Regex)
    - PaddedString() should allow providing a different padding character.
    - NullStripped() should not be a subconstruct that reads the entire stream instead it should be an
        Adapter that strips off the wrapped result.
    - Redefine String() to be the non-padded version of PaddedString()
        - (This will break compatibility with 2.8 but it seems to make the most sense)
        - Alternatively, rename PaddedString() back to String() and provide an flag parameter to determine if
          null characters should be stripped.
    - Add the path to ConstructError exceptions. This will greatly help with debugging.
    - Add deepcopy functionality for Container classes.
    - Embedding should also embed the context.
        - Also, Embedded should just be a function that toggles flagembedded instead of being it's own class.
    - remove _io from resulting Container objects after a parse. Doesn't look to be used for anything.
"""

from __future__ import absolute_import

from future.builtins import bytes, str

import codecs
import collections
import sys

import construct
import construct.core
import construct.debug
from construct import *
from construct.core import *


class Range(Subconstruct):
    r"""
    A homogenous array of elements. The array will iterate through between ``min`` to ``max`` times. If an exception occurs (EOF, validation error), the repeater exits cleanly. If less than ``min`` units have been successfully parsed, a RangeError is raised.

    .. seealso:: Analog :func:`~construct.core.GreedyRange` that parses until end of stream.

    .. note:: This object requires a seekable stream for parsing.

    :param min: the minimal count
    :param max: the maximal count
    :param subcon: the subcon to process individual elements

    Example::

        >>> Range(3, 5, Byte).build([1,2,3,4])
        '\x01\x02\x03\x04'
        >>> Range(3, 5, Byte).parse(_)
        ListContainer([1, 2, 3, 4])

        >>> Range(3, 5, Byte).build([1,2])
        Traceback (most recent call last):
            ...
        RangeError: expected from 3 to 5 elements, found 2
        >>> Range(3, 5, Byte).build([1,2,3,4,5,6])
        Traceback (most recent call last):
            ...
        RangeError: expected from 3 to 5 elements, found 6
    """
    __slots__ = ["min", "max"]

    def __init__(self, min, max, subcon):
        super(Range, self).__init__(subcon)
        self.min = min
        self.max = max

    def _parse(self, stream, context, path):
        min_ = evaluate(self.min, context)
        max_ = evaluate(self.max, context)
        if not 0 <= min_ <= max_ <= sys.maxsize:
            raise RangeError("[{}] unsane min {} and max {}".format(path, min_, max_))
        obj = ListContainer()
        try:
            i = 0
            while len(obj) < max_:
                context._index = i
                fallback = stream.tell()
                obj.append(self.subcon._parsereport(stream, context, path))
                if stream.tell() == fallback:
                    raise ExplicitError("[{}] Infinite loop detected.".format(path))
                i += 1
        except StopIteration:
            pass
        except ExplicitError:
            raise
        except Exception:  # TODO: catch ConstructError instead?
            if len(obj) < min_:
                raise RangeError("[{}] expected {} to {}, found {}".format(path, min_, max_, len(obj)))
            stream.seek(fallback)
        return obj

    def _build(self, obj, stream, context, path):
        min_ = evaluate(self.min, context)
        max_ = evaluate(self.max, context)
        if not 0 <= min_ <= max_ <= sys.maxsize:
            raise RangeError("[{}] unsane min {} and max {}".format(path, min_, max_))
        if not isinstance(obj, collections.Sequence):
            raise RangeError("[{}] expected sequence type, found {}".format(path, type(obj)))
        if not min_ <= len(obj) <= max_:
            raise RangeError("[{}] expected from {} to {} elements, found {}".format(path, min_, max_, len(obj)))
        retlist = ListContainer()
        try:
            for i, subobj in enumerate(obj):
                context._index = i
                buildret = self.subcon._build(subobj, stream, context, path)
                retlist.append(buildret)
        except StopIteration:
            pass
        except ExplicitError:
            raise
        except Exception:
            if len(obj) < min_:
                raise RangeError("[{}] expected {} to {}, found {}".format(path, min_, max_, len(obj)))
            else:
                raise
        return retlist

    def _sizeof(self, context, path):
        # WARNING: possibly broken by StopIf
        try:
            min_ = evaluate(self.min, context)
            max_ = evaluate(self.max, context)
        except (KeyError, AttributeError):
            raise SizeofError("cannot calculate size, key not found in context")
        if min_ == max_:
            return min_ * self.subcon._sizeof(context, path)
        else:
            raise SizeofError("cannot calculate size")


def CString(encoding='utf-8'):
    r"""
    Adds default encoding option to CString().

    >>> CString().parse(b'hello\x00')
    u'hello'
    >>> CString().parse(b'hello\x00\xff\xff')
    u'hello'
    >>> CString(encoding='utf-16').parse(b'\xff\xfeh\x00e\x00l\x00l\x00o\x00\x00\x00')  # FFFE is BOM for utf-16-le
    u'hello'
    >>> CString(encoding='utf-16').parse(b'h\x00e\x00l\x00l\x00o\x00\x00\x00')
    u'hello'
    >>> CString(encoding='utf-16').build(u'hello')
    '\xff\xfeh\x00e\x00l\x00l\x00o\x00\x00\x00'
    >>> CString(encoding='utf-32').build(u'hello')
    '\xff\xfe\x00\x00h\x00\x00\x00e\x00\x00\x00l\x00\x00\x00l\x00\x00\x00o\x00\x00\x00\x00\x00\x00\x00'

    Make sure to specify 'le' or 'be' in the encoding if you don't want BOM markers when building.
    >>> CString(encoding='utf-32-le').build(u'hello')
    'h\x00\x00\x00e\x00\x00\x00l\x00\x00\x00l\x00\x00\x00o\x00\x00\x00\x00\x00\x00\x00'
    >>> CString(encoding='utf-32-be').build(u'hello')
    '\x00\x00\x00h\x00\x00\x00e\x00\x00\x00l\x00\x00\x00l\x00\x00\x00o\x00\x00\x00\x00'
    """
    return construct.CString(encoding)


def PaddedString(length, encoding='utf-8'):
    r"""
    Adds default encoding option to PaddedString().

    NOTE: When using this to build a multi-byte encoded string you need to be aware of the extra space that can be taken
    up by BOM markings when specifying the length.
    If you don't want BOM. Make sure to explicitly specify "le" or "be" at the end of your encoding.
    >>> u'hi'.encode('utf-16')
    '\xff\xfeh\x00i\x00'
    >>> u'hi'.encode('utf-16-le')
    'h\x00i\x00'

    :param length: length in bytes (not unicode characters), as int or context function
    :param encoding: encoding (e.g. "utf8") or None for bytes

    e.g.
    >>> StringEncoded(Bytes(10), 'utf-16').parse(b'h\x00e\x00l\x00l\x00o\x00')
    u'hello'
    >>> PaddedString(10, encoding='utf-16').parse(b'h\x00e\x00l\x00l\x00o\x00')
    u'hello'
    >>> PaddedString(12, encoding='utf-16').build(u'hello')
    '\xff\xfeh\x00e\x00l\x00l\x00o\x00'
    >>> PaddedString(10, encoding='utf-16le').build(u'hello')
    'h\x00e\x00l\x00l\x00o\x00'
    >>> PaddedString(16, encoding='utf-16le').build(u'hello')
    'h\x00e\x00l\x00l\x00o\x00\x00\x00\x00\x00\x00\x00'
    >>> PaddedString(16, encoding='utf-16').parse(b'h\x00e\x00l\x00l\x00o\x00\x00\x00\x00\x00\x00\x00')
    u'hello'

    Works with utf-32 in the same way.
    >>> PaddedString(20, encoding='utf-32-le').build(u'hello')
    'h\x00\x00\x00e\x00\x00\x00l\x00\x00\x00l\x00\x00\x00o\x00\x00\x00'
    >>> PaddedString(20, encoding='utf-32').parse(b'h\x00\x00\x00e\x00\x00\x00l\x00\x00\x00l\x00\x00\x00o\x00\x00\x00')
    u'hello'

    Also, still works with regular single byte encodings.
    >>> PaddedString(5).build(u'hello')
    'hello'
    >>> PaddedString(5).parse(b'hello')
    u'hello'

    A StringError (type of ConstructError) will be raised if the string cannot be decoded with the given encoding.
    >>> PaddedString(8).parse(b'hello\x00\xff\xff')
    Traceback (most recent call last):
        ...
    StringError: string decoding failed: 'utf8' codec can't decode byte 0xff in position 6: invalid start byte
    """
    return construct.PaddedString(length, encoding)

# Alias for original 2.8 name
# FIXME: String() should not remove the null padding!
String = PaddedString


def GreedyString(encoding='utf-8'):
    """Adds default encoding option to PaddedString()."""
    return construct.GreedyString(encoding)


class Compressed(Adapter):
    r"""
    Replaces the original Compressed construct to improve functionality:
        - supports providing a custom encoding module or object.
            - (provide any object that has a "decompress" and "compress" function in the lib parameter.)
        - produces a ConstructError if compressed/decompression fails.
            - (You can turn this off by setting wrap_exception=False)
        - uses Adapter instead of Tunnel in order to allow it be embedded within other constructs.
            - (Original one read entire stream, no matter the subcon you provide.)

    e.g.
    >>> import zlib
    >>> Compressed(GreedyBytes, zlib).build('hello world')
    'x\x9c\xcbH\xcd\xc9\xc9W(\xcf/\xcaI\x01\x00\x1a\x0b\x04]'
    >>> Compressed(GreedyBytes, zlib).parse(_)
    'hello world'
    >>> import dc3cipher
    >>> lzma = dc3cipher.new('lzma')
    >>> Compressed(GreedyBytes, lzma).build('hello world')
    ']\x00\x00\x80\x00\x004\x19I\xee\x8d\xe9\x17\x89:3`\x05\xf7\xcfd\xff\xfbx \x00'
    >>> Compressed(GreedyBytes, lzma).parse(_)
    'hello world'

    Now that this is an Adapter, it can be become part of a larger struct.
    >>> spec = Struct(
    ...     'magic' / Const('YUP'),
    ...     'data' / Compressed(Bytes(26), lzma),
    ...     'trailer' / Int32ul,
    ... )
    >>> spec.parse('YUP]\x00\x00\x80\x00\x004\x19I\xee\x8d\xe9\x17\x89:3`\x05\xf7\xcfd\xff\xfbx \x00\x03\x00\x00\x00')
    Container(magic='YUP')(data='hello world')(trailer=3)
    >>> spec.build(_)
    'YUP]\x00\x00\x80\x00\x004\x19I\xee\x8d\xe9\x17\x89:3`\x05\xf7\xcfd\xff\xfbx \x00\x03\x00\x00\x00'
    """
    __slots__ = ["lib", "wrap_exception"]

    def __init__(self, subcon, lib, wrap_exception=True):
        super(Compressed, self).__init__(subcon)
        self.wrap_exception = wrap_exception
        if hasattr(lib, "compress") and hasattr(lib, "decompress"):
            self.lib = lib
        elif lib == "zlib":
            import zlib
            self.lib = zlib
        elif lib == "gzip":
            import gzip
            self.lib = gzip
        elif lib == "bzip2":
            import bz2
            self.lib = bz2
        else:
            raise ValueError('Invalid lib parameter: {}'.format(lib))

    def _decode(self, data, context, path):
        try:
            return self.lib.decompress(data)
        except Exception as e:
            if self.wrap_exception:
                raise ConstructError('Decompression failed with error: {}'.format(e))
            else:
                raise

    def _encode(self, data, context, path):
        try:
            return self.lib.compress(data)
        except Exception as e:
            if self.wrap_exception:
                raise ConstructError('Compression failed with error: {}'.format(e))
            else:
                raise


class Union(construct.Union):
    """
    Patches the Union() Construct to not require the parsefrom parameter. (defaults to None)
    """

    def __init__(self, parsefrom_or_subcon, *subcons, **subconskw):
        if isinstance(parsefrom_or_subcon, Construct):
            parsefrom = None
            subcons = (parsefrom_or_subcon,) + subcons
        else:
            parsefrom = parsefrom_or_subcon
        super(Union, self).__init__(parsefrom, *subcons, **subconskw)


# Map an integer in the inclusive range 0-255 to its string byte representation
PRINTABLE = [bytes2str(int2byte(i)) if 32 <= i < 128 else '.' for i in range(256)]
HEXPRINT = [format(i, '02X') for i in range(256)]

# Copy of construct.lib.hex.hexdump but removes the "hexundump(" string.
# Not sure why that was added....
def hexdump(data, linesize):
    r"""
    Turns bytes into a unicode string of the format:

    ::

        >>> print(hexdump(b'0' * 100, 16))
        hexundump(\"\"\"
        0000   30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30   0000000000000000
        0010   30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30   0000000000000000
        0020   30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30   0000000000000000
        0030   30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30   0000000000000000
        0040   30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30   0000000000000000
        0050   30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30   0000000000000000
        0060   30 30 30 30                                       0000
        \"\"\")
    """
    if len(data) < 16**4:
        fmt = "%%04X   %%-%ds   %%s" % (3*linesize-1,)
    elif len(data) < 16**8:
        fmt = "%%08X   %%-%ds   %%s" % (3*linesize-1,)
    else:
        raise ValueError("hexdump cannot process more than 16**8 or 4294967296 bytes")
    prettylines = []
    for i in range(0, len(data), linesize):
        line = data[i:i+linesize]
        hextext = " ".join(HEXPRINT[b] for b in iterateints(line))
        rawtext = "".join(PRINTABLE[b] for b in iterateints(line))
        prettylines.append(fmt % (i, str(hextext), str(rawtext)))
    return "\n".join(prettylines)


class Probe(construct.Probe):
    """
    Patches back some of the features of Probe() that were removed:

        - The lookahead stream is enabled by default
        - Use hexdump instead of hexlify to display lookahead stream
        - Allows for setting a name
    """
    def __init__(self, into=None, lookahead=128, name=None):
        self.print_name = name
        super(Probe, self).__init__(into=into, lookahead=lookahead)

    def printout(self, stream, context, path):
        print("--------------------------------------------------")
        print("Probe {}".format(self.print_name or ''))
        print("Path: {}".format(path))
        if self.into:
            print("Into: {!r}".format(self.into))

        if self.lookahead and stream is not None:
            fallback = stream.tell()
            stream_bytes = stream.read(self.lookahead)
            stream.seek(fallback)
            if stream_bytes:
                print("Stream peek:\n{}".format(hexdump(stream_bytes, 32)))
            else:
                print("Stream peek: EOF reached")

        if context is not None:
            if self.into:
                try:
                    subcontext = self.into(context)
                    print(subcontext)
                except Exception:
                    print("Failed to compute {!r} on the context {!r}".format(self.into, context))
            else:
                print(context)
        print("--------------------------------------------------")


class FocusedSeq(construct.FocusedSeq):
    """
    Patches FocusedSeq to add back support for supplying an index.

    Also fixes the build and parse functions.
    """

    def _parse(self, stream, context, path):
        context = Container(_ = context, _params = context._params, _root = None, _parsing = context._parsing, _building = context._building, _sizing = context._sizing, _subcons = self._subcons, _io = stream, _index = context.get("_index", None))
        context._root = context._.get("_root", context)
        parsebuildfrom = evaluate(self.parsebuildfrom, context)

        found = False  # Must use separate flag because returning a parse result of None is valid.
        finalret = None
        for i, sc in enumerate(self.subcons):
            parseret = sc._parsereport(stream, context, path)
            context[i] = parseret  # PATCH: re-added ability to reference by index.
            if sc.name:
                context[sc.name] = parseret
            if sc.name == parsebuildfrom or i == parsebuildfrom:
                finalret = parseret
                found = True

        if not found:
            raise ConstructError("Unable to find entry: {}".format(parsebuildfrom))

        return finalret

    def _build(self, obj, stream, context, path):
        context = Container(_ = context, _params = context._params, _root = None, _parsing = context._parsing, _building = context._building, _sizing = context._sizing, _subcons = self._subcons, _io = stream, _index = context.get("_index", None))
        context._root = context._.get("_root", context)
        parsebuildfrom = evaluate(self.parsebuildfrom, context)

        found = False
        finalret = None
        for i, sc in enumerate(self.subcons):
            if sc.name == parsebuildfrom or i == parsebuildfrom:
                sub_obj = obj
            else:
                sub_obj = context._.get(sc.name, context._.get(i, None))
            try:
                buildret = sc._build(sub_obj, stream, context, path)
            except ConstructError as e:
                raise ConstructError("Unable to build field at index: {}".format(i))

            context[i] = buildret
            if sc.name:
                context[sc.name] = buildret

            if sc.name == parsebuildfrom or i == parsebuildfrom:
                finalret = buildret
                found = True

        if not found:
            raise ConstructError("Unable to find entry: {}".format(parsebuildfrom))

        return finalret


class Mapping(construct.Mapping):
    r"""
    Patches Mapping to allow non-symmetric mappings by swapping the mapping
    from encoding (building) to decoding (parsing) and allow for an optional mapping in the other direction.

    :param subcon: Construct instance
    :param mapping: dict, for decoding (parsing) mapping
    :param enc_mapping: Optional mapping for encoding (building), otherwise the reversed decoding mapping it used

    Example::
        >>> spec = Mapping(Byte, {0: u'a', 1: u'b', 2: u'b'})
        >>> spec.parse(b'\x02')
        u'b'

        # Reverse mapping is sorted so 1 will be used instead of 2.
        >>> spec.build(u'b')
        '\x01'
    """

    def __init__(self, subcon, dec_mapping, enc_mapping=None):
        super(Mapping, self).__init__(subcon, {})
        self.decmapping = dec_mapping
        self.encmapping = enc_mapping or {v: k for k, v in sorted(dec_mapping.items(), reverse=True)}



def _patch_pop():
    """
    Patches the pop() function in Container to allow for a default value.
    """
    def pop(self, key, *default):
        try:
            val = dict.pop(self, key, *default)
            self.__keys_order__.remove(key)
            return val
        except ValueError:
            if default:
                return default[0]
            else:
                raise KeyError

    Container.pop = pop


def _patch_slice():
    """Patches the slicing mechanism to use Range"""
    orig_get_item = Construct.__getitem__

    def __getitem__(self, count):
        if isinstance(count, slice):
            if count.step is not None:
                raise ValueError("slice must not contain a step: %r" % count)
            min = 0 if count.start is None else count.start
            max = sys.maxsize if count.stop is None else count.stop
            return Range(min, max, self)
        else:
            return orig_get_item(self, count)

    Construct.__getitem__ = __getitem__


def _patch_StringEncoded():
    """
    Patches StringEncoded to throw a ConstructError type exception if decoding fails.

    Fixes: github.com/construct/construct/issues/743
    """
    orig_decode = construct.StringEncoded._decode

    def _decode(self, obj, context, path):
        try:
            return orig_decode(self, obj, context, path)
        except UnicodeDecodeError as e:
            raise StringError("[{}] string decoding failed: {}".format(path, e))

    construct.StringEncoded._decode = _decode


def _patch_sizeof():
    """
    Patches the sizeof() function in Struct, Sequence, and FocusedSeq to properly provide context.

    Fixes: github.com/construct/construct/issues/771
    """
    def _sizeof(self, context, path):
        # Removed the context manipulation.
        try:
            # Added back dereferencing nested context that was incorrectly removed.
            def isStruct(sc):
                return isStruct(sc.subcon) if isinstance(sc, Renamed) else isinstance(sc, Struct)
            def nest(context, sc):
                if isStruct(sc) and not sc.flagembedded and sc.name in context:
                    context2 = context[sc.name]
                    context2["_"] = context
                    return context2
                else:
                    return context
            return sum(sc._sizeof(nest(context, sc), path) for sc in self.subcons)
        except (KeyError, AttributeError):
            raise SizeofError("cannot calculate size, key not found in context")

    # Conveniently, all 3 Constructs are implemented in the same way.
    construct.Struct._sizeof = _sizeof
    construct.Sequence._sizeof = _sizeof
    construct.FocusedSeq._sizeof = _sizeof


def _patch_encodingunit():
    """
    Patches the encodingunit() function that is used to calculate
    sizes for null terminated strings.

    This fixes the limitation of having a hardcoded set of supported encodings found in the original
    implementation.
    """
    # must be ordered largest to smallest
    _BOM_BYTES = (
        codecs.BOM_UTF32_LE,
        codecs.BOM_UTF32_BE,
        codecs.BOM_UTF16_LE,
        codecs.BOM_UTF16_BE,
        codecs.BOM_UTF8,
    )

    # NOTE: We can't patch in our version of encodingunit() so we are going to have to reimplement
    # the functions that use it (seen below)
    def encodingunit(encoding):
        r"""
        >>> encodingunit('utf-8')
        b'\x00'
        >>> encodingunit('utf-16le')
        b'\x00\x00'
        >>> encodingunit('utf-16')
        b'\x00\x00'
        >>> encodingunit('utf-32')
        b'\x00\x00\x00\x00'
        >>> encodingunit('cp950')
        b'\x00'
        """
        # Check "basic" byte size without BOM mark
        encoding = encoding.lower()
        encoded = u'\0'.encode(encoding)
        for bom_bytes in _BOM_BYTES:
            if encoded.startswith(bom_bytes) and len(bom_bytes) < len(encoded):
                encoded = encoded[len(bom_bytes):]
                break
        return bytes(len(encoded))

    construct.core.encodingunit = encodingunit


def _patch_mergefields():
    """
    Patches the mergefields() function to remove the hardcoded list of embeddable classes.

    This fixes the issue of trying to wrap Embedded around a Bitwise component.

    Fixes: github.com/construct/construct/issues/TODO
    """
    def mergefields(*subcons):
        def select(sc):
            # If it quacks like a duck...
            if hasattr(sc, 'subcons'):
                return sc.subcons
            elif hasattr(sc, 'subcon'):
                return select(sc.subcon)
            raise ConstructError(
                "Embedding only works with: Struct Sequence FocusedSeq Union LazyStruct: {!r}".format(sc))

        result = []
        for sc in subcons:
            if sc.flagembedded:
                result.extend(select(sc))
            else:
                result.append(sc)
        return result

    construct.core.mergefields = mergefields


def _patch():
    """Patches 2.9 with 2.8 features and other general fixes."""
    _patch_pop()
    _patch_slice()
    _patch_StringEncoded()
    _patch_sizeof()
    _patch_encodingunit()
    _patch_mergefields()


_patch()



"""
A central location to store common windows enumerations.
This module will be imported along with 'from mwcp.utils import construct'
"""

from __future__ import absolute_import, division

import datetime

from . import version28 as construct
from .version28 import this, len_

from . import network, datetime_, windows_enums
from .windows_constants import *


"""PEFILE STRUCTURES"""

IMAGE_DOS_HEADER = construct.Struct(
    "e_magic" / construct.Default(construct.Bytes(2), b"MZ"),
    "e_cblp" / construct.Int16ul,
    "e_cp" / construct.Int16ul,
    "e_crlc" / construct.Int16ul,
    "e_cparhdr" / construct.Int16ul,
    "e_mimalloc" / construct.Int16ul,
    "e_maxalloc" / construct.Int16ul,
    "e_ss" / construct.Int16ul,
    "e_sp" / construct.Int16ul,
    "e_csum" / construct.Int16ul,
    "e_ip" / construct.Int16ul,
    "e_cs" / construct.Int16ul,
    "e_lfarlc" / construct.Int16ul,
    "e_ovno" / construct.Int16ul,
    "e_res1" / construct.Bytes(8),
    "e_oemid" / construct.Int16ul,
    "e_oeminfo" / construct.Int16ul,
    "e_res2" / construct.Bytes(20),
    "e_lfanew" / construct.Int32ul
)


IMAGE_SECTION_HEADER = construct.Struct(
    "Name" / construct.String(8),
    "VirtualSize" / construct.Int32ul,  # alias "PhysicalAddress"
    "VirtualAddress" / construct.Int32ul,
    "SizeOfRawData" / construct.Int32ul,
    "PointerToRawData" / construct.Int32ul,
    "PointerToRelocations" / construct.Default(construct.Int32ul, 0),
    "PointerToLinenumbers" / construct.Default(construct.Int32ul, 0),
    "NumberOfRelocations" / construct.Default(construct.Int16ul, 0),
    "NumberOfLinenumbers" / construct.Default(construct.Int16ul, 0),
    "Characteristics" / construct.FlagsEnum(
        construct.Int32ul,
        IMAGE_SCN_TYPE_NO_PAD=0x00000008,
        IMAGE_SCN_CNT_CODE=0x00000020,
        IMAGE_SCN_CNT_INITIALIZED_DATA=0x00000040,
        IMAGE_SCN_CNT_UNINITIALIZED_DATA=0x00000080,
        IMAGE_SCN_LNK_OTHER=0x00000100,
        IMAGE_SCN_LNK_INFO=0x00000200,
        IMAGE_SCN_LNK_REMOVE=0x00000800,
        IMAGE_SCN_LNK_COMDAT=0x00001000,
        IMAGE_SCN_NO_DEFER_SPEC_EXC=0x00004000,
        IMAGE_SCN_GPREL=0x00008000,
        IMAGE_SCN_MEM_PURGEABLE=0x00020000,
        IMAGE_SCN_MEM_LOCKED=0x00040000,
        IMAGE_SCN_MEM_PRELOAD=0x00080000,
        IMAGE_SCN_ALIGN_1BYTES=0x00100000,
        IMAGE_SCN_ALIGN_2BYTES=0x00200000,
        IMAGE_SCN_ALIGN_4BYTES=0x00300000,
        IMAGE_SCN_ALIGN_8BYTES=0x00400000,
        IMAGE_SCN_ALIGN_16BYTES=0x00500000,
        IMAGE_SCN_ALIGN_32BYTES=0x00600000,
        IMAGE_SCN_ALIGN_64BYTES=0x00700000,
        IMAGE_SCN_ALIGN_128BYTES=0x00800000,
        IMAGE_SCN_ALIGN_256BYTES=0x00900000,
        IMAGE_SCN_ALIGN_512BYTES=0x00A00000,
        IMAGE_SCN_ALIGN_1024BYTES=0x00B00000,
        IMAGE_SCN_ALIGN_2048BYTES=0x00C00000,
        IMAGE_SCN_ALIGN_4096BYTES=0x00D00000,
        IMAGE_SCN_ALIGN_8192BYTES=0x00E00000,
        IMAGE_SCN_LNK_NRELOC_OVFL=0x01000000,
        IMAGE_SCN_MEM_DISCARDABLE=0x02000000,
        IMAGE_SCN_MEM_NOT_CACHED=0x04000000,
        IMAGE_SCN_MEM_NOT_PAGED=0x08000000,
        IMAGE_SCN_MEM_SHARED=0x10000000,
        IMAGE_SCN_MEM_EXECUTE=0x20000000,
        IMAGE_SCN_MEM_READ=0x40000000,
        IMAGE_SCN_MEM_WRITE=0x80000000,
    )
)

IMAGE_DATA_DIRECTORY = construct.Struct(
    "VirtualAddress" / construct.Int32ul,
    "Size" / construct.Int32ul,
)

IMAGE_EXPORT_DIRECTORY = construct.Struct(
    "Characteristics" / construct.Default(construct.Int32ul, 0),
    "TimeDateStamp" / datetime_.EpochTime,
    "MajorVersion" / construct.Int16ul,
    "MinorVersion" / construct.Int16ul,
    "Name" / construct.Int32ul,  # rva pointer to the name
    "Base" / construct.Int32ul,
    "NumberOfFunctions" / construct.Int32ul,
    "NumberOfNames" / construct.Int32ul,
    "AddressOfFunctions" / construct.Int32ul,
    "AddressOfNames" / construct.Int32ul,
    "AddressOfNameOrdinals" / construct.Int32ul,
)

IMAGE_IMPORT_DESCRIPTOR = construct.Struct(
    "Characteristics" / construct.Int32ul,
    "TimeDateStamp" / construct.Int32ul,
    "ForwarderChain" / construct.Int32ul,
    "Name" / construct.Int32ul,  # rva pointer to the name
    "FirstThunk" / construct.Int32ul,
)

IMAGE_OPTIONAL_HEADER = construct.Struct(
    "Magic" / construct.OneOf(construct.Int16ul, [
        IMAGE_NT_OPTIONAL_HDR32_MAGIC, IMAGE_NT_OPTIONAL_HDR64_MAGIC, IMAGE_ROM_OPTIONAL_HDR_MAGIC]),
    "MajorLinkerVersion" / construct.Byte,
    "MinorLinkerVersion" / construct.Byte,
    "SizeOfCode" / construct.Int32ul,
    "SizeOfInitializedData" / construct.Int32ul,
    "SizeOfUninitializedData" / construct.Int32ul,
    "AddressOfEntryPoint" / construct.Int32ul,
    "BaseOfCode" / construct.Int32ul,
    "BaseOfData" / construct.If(this.Magic != IMAGE_NT_OPTIONAL_HDR64_MAGIC, construct.Int32ul),
    "ImageBase" / construct.IfThenElse(
        this.Magic == IMAGE_NT_OPTIONAL_HDR64_MAGIC, construct.Int64ul, construct.Int32ul
    ),
    "SectionAlignment" / construct.Int32ul,
    "FileAlignment" / construct.Int32ul,
    "MajorOperatingSystemVersion" / construct.Int16ul,
    "MinorOperatingSystemVersion" / construct.Int16ul,
    "MajorImageVersion" / construct.Int16ul,
    "MinorImageVersion" / construct.Int16ul,
    "MajorSubsystemVersion" / construct.Int16ul,
    "MinorSubsystemVersion" / construct.Int16ul,
    "Win32VersionValue" / construct.Default(construct.Int32ul, 0),  # must be 0
    "SizeOfImage" / construct.Int32ul,
    "SizeOfHeaders" / construct.Int32ul,
    "CheckSum" / construct.Int32ul,
    # TODO: Use enums instead?
    "Subsystem" / construct.OneOf(construct.Int16ul, [
        IMAGE_SUBSYSTEM_UNKNOWN,
        IMAGE_SUBSYSTEM_NATIVE,
        IMAGE_SUBSYSTEM_WINDOWS_GUI,
        IMAGE_SUBSYSTEM_WINDOWS_CUI,
        IMAGE_SUBSYSTEM_OS2_CUI,
        IMAGE_SUBSYSTEM_POSIX_CUI,
        IMAGE_SUBSYSTEM_WINDOWS_CE_GUI,
        IMAGE_SUBSYSTEM_EFI_APPLICATION,
        IMAGE_SUBSYSTEM_EFI_BOOT_SERVICE_DRIVER,
        IMAGE_SUBSYSTEM_EFI_RUNTIME_DRIVER,
        IMAGE_SUBSYSTEM_EFI_ROM,
        IMAGE_SUBSYSTEM_XBOX,
        IMAGE_SUBSYSTEM_WINDOWS_BOOT_APPLICATION,
    ]),
    "DllCharacteristics" / construct.FlagsEnum(
        construct.Int16ul,
        IMAGE_DLLCHARACTERISTICS_HIGH_ENTROPY_VA=0x0020,
        IMAGE_DLLCHARACTERISTICS_DYNAMIC_BASE=0x0040,
        IMAGE_DLLCHARACTERISTICS_FORCE_INTEGRITY=0x0080,
        IMAGE_DLLCHARACTERISTICS_NX_COMPAT=0x0100,
        IMAGE_DLLCHARACTERISTICS_NO_ISOLATION=0x0200,
        IMAGE_DLLCHARACTERISTICS_NO_SEH=0x0400,
        IMAGE_DLLCHARACTERISTICS_NO_BIND=0x0800,
        IMAGE_DLLCHARACTERISTICS_APPCONTAINER=0x1000,
        IMAGE_DLLCHARACTERISTICS_WDM_DRIVER=0x2000,
        IMAGE_DLLCHARACTERISTICS_GUARD_CF=0x4000,
        IMAGE_DLLCHARACTERISTICS_TERMINAL_SERVER_AWARE=0x8000,
    ),
    "SizeOfStackReserve" / construct.IfThenElse(
        this.Magic == IMAGE_NT_OPTIONAL_HDR64_MAGIC, construct.Int64ul, construct.Int32ul
    ),
    "SizeOfStackCommit" / construct.IfThenElse(
        this.Magic == IMAGE_NT_OPTIONAL_HDR64_MAGIC, construct.Int64ul, construct.Int32ul
    ),
    "SizeOfHeapReserve" / construct.IfThenElse(
        this.Magic == IMAGE_NT_OPTIONAL_HDR64_MAGIC, construct.Int64ul, construct.Int32ul
    ),
    "SizeOfHeapCommit" / construct.IfThenElse(
        this.Magic == IMAGE_NT_OPTIONAL_HDR64_MAGIC, construct.Int64ul, construct.Int32ul
    ),
    "LoaderFlags" / construct.Int32ul,
    "NumberOfRvaAndSizes" / construct.Rebuild(construct.Int32ul, construct.len_(this.DataDirectory)),
    "DataDirectory" / construct.Default(IMAGE_DATA_DIRECTORY[this.NumberOfRvaAndSizes], DEFAULT_DATA_DIRECTORIES[:]),
)

IMAGE_FILE_HEADER = construct.Struct(
    "Machine" / construct.Int16ul,  # IMAGE_FILE_MACHINE_*
    "NumberOfSections" / construct.Int16ul,
    "TimeDateStamp" / construct.Int32ul,
    "PointerToSymbolTable" / construct.Default(construct.Int32ul, 0),
    "NumberOfSymbols" / construct.Default(construct.Int32ul, 0),
    # NOTE: This defaults to assuming a 32-bit PE when building if the SizeOfOptionalHeader isn't provided in the context.
    "SizeOfOptionalHeader" / construct.Default(
        construct.Int16ul, IMAGE_OPTIONAL_HEADER.sizeof(Magic=IMAGE_NT_OPTIONAL_HDR32_MAGIC, NumberOfRvaAndSizes=16)),
    "Characteristics" / construct.FlagsEnum(
        construct.Int16ul,
        IMAGE_FILE_RELOCS_STRIPPED=0x0001,
        IMAGE_FILE_EXECUTABLE_IMAGE=0x0002,
        IMAGE_FILE_LINE_NUMS_STRIPPED=0x0004,
        IMAGE_FILE_LOCAL_SYMS_STRIPPED=0x0008,
        IMAGE_FILE_AGGRESIVE_WS_TRIM=0x0010,
        IMAGE_FILE_LARGE_ADDRESS_AWARE=0x0020,
        IMAGE_FILE_BYTES_REVERSED_LO=0x0080,
        IMAGE_FILE_32BIT_MACHINE=0x0100,
        IMAGE_FILE_DEBUG_STRIPPED=0x0200,
        IMAGE_FILE_REMOVABLE_RUN_FROM_SWAP=0x0400,
        IMAGE_FILE_NET_RUN_FROM_SWAP=0x0800,
        IMAGE_FILE_SYSTEM=0x1000,
        IMAGE_FILE_DLL=0x2000,
        IMAGE_FILE_UP_SYSTEM_ONLY=0x4000,
        IMAGE_FILE_BYTES_REVERSED_HI=0x8000,
    ),
)

IMAGE_NT_HEADERS = construct.Struct(
    "Signature" / construct.Default(construct.Int32ul, 0x4550),  # b'PE\x00\x00'
    "FileHeader" / IMAGE_FILE_HEADER,
    "OptionalHeader" / IMAGE_OPTIONAL_HEADER
)

PEFILE_HEADER = construct.Struct(
    "DosHeader" / IMAGE_DOS_HEADER,
    # TODO: Use construct.FixedSized() if we ever update construct.
    "DosStub" / construct.Bytes(this.DosHeader.e_lfanew - IMAGE_DOS_HEADER.sizeof()),
    "NTHeaders" / IMAGE_NT_HEADERS,
    "SectionTable" / IMAGE_SECTION_HEADER[this.NTHeaders.FileHeader.NumberOfSections],
)

"""WINSOCK STRUCTURES"""

SOCKADDR_IN = construct.Struct(
    "sin_family" / construct.Int16ul,
    "sin_port" / construct.Int16ub,  # in network byte order
    "sin_addr" / network.IP4Address,
    "sin_zero" / construct.Bytes(8)
)

# Same as SOCKADDR_IN but with the port as little endian.
SOCKADDR_IN_L = construct.Struct(
    "sin_family" / construct.Int16ul,
    "sin_port" / construct.Int16ul,
    "sin_addr" / network.IP4Address,
    "sin_zero" / construct.Bytes(8)
)

"""CRYPTO STRUCTURES"""

PUBLICKEYSTRUC = construct.Struct(
    "type" / construct.Byte,
    "version" / construct.Byte,
    "reserved" / construct.Int16ul,
    "algid" / windows_enums.AlgorithmID(construct.Int32ul),
)

PUBLICKEYBLOB = construct.Struct(
    "publickeystruc" / PUBLICKEYSTRUC,
    construct.Check(this.publickeystruc.algid == "CALG_RSA_KEYX"),
    construct.Const(b"RSA1"),
    "bitlen" / construct.Int32ul,
    construct.Check((this.bitlen % 8) == 0),
    "pubexponent" / construct.Int32ul,
    "modulus" / construct.BytesInteger(this.bitlen // 8, swapped=True)
)

PRIVATEKEYBLOB = construct.Struct(
    "publickeystruc" / PUBLICKEYSTRUC,
    construct.Check(this.publickeystruc.algid == "CALG_RSA_KEYX"),
    construct.Const(b"RSA2"),
    "bitlen" / construct.Int32ul,
    construct.Check((this.bitlen % 8) == 0),
    "pubexponent" / construct.Int32ul,
    "modulus" / construct.BytesInteger(this.bitlen // 8, swapped=True),
    "P" / construct.BytesInteger(this.bitlen // 16, swapped=True),
    "Q" / construct.BytesInteger(this.bitlen // 16, swapped=True),
    # d % (p - 1)
    "Dp" / construct.BytesInteger(this.bitlen // 16, swapped=True),
    # d % (q - 1)
    "Dq" / construct.BytesInteger(this.bitlen // 16, swapped=True),
    # ~(q % p)
    "Iq" / construct.BytesInteger(this.bitlen // 16, swapped=True),
    # Private Exponent
    "D" / construct.BytesInteger(this.bitlen // 8, swapped=True)
)

"""TIME STRUCTURES"""

SYSTEMTIME = construct.Struct(
    "wYear" / construct.Int16ul,
    "wMonth" / construct.Int16ul,
    "wDayOfWeek" / construct.Int16ul,
    "wDay" / construct.Int16ul,
    "wHour" / construct.Int16ul,
    "wMinute" / construct.Int16ul,
    "wSecond" / construct.Int16ul,
    "wMilliseconds" / construct.Int16ul,
)


# TODO: Implement _encode
class _SystemTimeAdapter(construct.Adapter):
    r"""
    Adapter to convert SYSTEMTIME structured data to datetime.datetime ISO format.

    >>> _SystemTimeAdapter(SYSTEMTIME).parse('\xdd\x07\t\x00\x03\x00\x12\x00\t\x00.\x00\x15\x00\xf2\x02')
    '2013-09-18T09:46:21.754000'
    """
    def _decode(self, obj, context, path):
        return datetime.datetime(
            obj.wYear, obj.wMonth, obj.wDay, obj.wHour, obj.wMinute, obj.wSecond, obj.wMilliseconds * 1000
        ).isoformat()

# Hide the adapter
SystemTime = _SystemTimeAdapter(SYSTEMTIME)


EPOCH_AS_FILETIME = 116444736000000000
HUNDREDS_OF_NANOSECONDS = 10000000


# TODO: Implement _encode
class _FileTimeAdapter(construct.Adapter):
    r"""
    Adapter to convert FILETIME structured data to datetime.datetime ISO format.
    Technically FILETIME is two 32-bit integers as dwLowDateTime and dwHighDateTime, but there is no need to do that

    >>> _FileTimeAdapter(construct.Int64ul).parse('\x00\x93\xcc\x11\xa7\x88\xd0\x01')
    '2015-05-07T05:20:33'
    """
    def _decode(self, obj, context, path):
        return datetime.datetime.fromtimestamp((obj - EPOCH_AS_FILETIME) / HUNDREDS_OF_NANOSECONDS).isoformat()

# Hide the adapter
FileTime = _FileTimeAdapter(construct.Int64ul)



"""Construct helpers for .NET"""

from .version28 import *


class _DotNetUInt(Construct):
    r"""
    DotNet encoded unsigned 32-bit integer, where first byte indicates the length of the integer.

    Example:

        >>> DotNetUInt.build(16)
        '\x10'
        >>> DotNetUInt.parse(_)
        16
        >>> DotNetUInt.build(256)
        '\x81\x00'
        >>> DotNetUInt.parse(_)
        256
        >>> DotNetUInt.build(0xffff)
        '\xc0\x00\xff\xff'
        >>> DotNetUInt.parse(_)
        65535
    """
    def _parse(self, stream, context, path):
        b = byte2int(stream_read(stream, 1))
        if b & 0x80 == 0:
            num = b
        elif b & 0xc0 == 0x80:
            num = ((b & 0x3f) << 8) + byte2int(stream_read(stream, 1))
        elif b & 0xe0 == 0xc0:
            num = (b & 0x1f) << 24
            num += byte2int(stream_read(stream, 1)) << 16
            num += byte2int(stream_read(stream, 1)) << 8
            num += byte2int(stream_read(stream, 1))
        else:
            raise ConstructError('DotNetUInt encountered an invalid string')
        return num

    def _build(self, obj, stream, context, path):
        if obj < 0:
            raise ConstructError("DotNetUInt cannot build from negative number")
        if obj > 0x1fffffff:
            raise ConstructError("DotNetUInt encountered too large a number")
        if obj < 0x80:
            stream_write(stream, int2byte(obj), 1)
        elif obj < 0x3fff:
            stream_write(stream, int2byte((obj >> 8) | 0x80), 1)
            stream_write(stream, int2byte(obj & 0xff), 1)
        else:
            stream_write(stream, int2byte((obj >> 24) | 0xc0), 1)
            stream_write(stream, int2byte((obj >> 16) & 0xff), 1)
            stream_write(stream, int2byte((obj >> 8) & 0xff), 1)
            stream_write(stream, int2byte(obj & 0xff), 1)


# using the @singleton decorator seems to break our ability to run doctests.
DotNetUInt = _DotNetUInt()


class _DotNetNullString(Construct):
    r"""
    DotNet null string, different from an empty zero-byte string, encoded as a single 0xff byte.

    Example:

        >>> repr(DotNetNullString.parse('\xff'))
        'None'
        >>> DotNetNullString.build(None)
        '\xff'
    """
    def _parse(self, stream, context, path):
        if stream_read(stream, 1) != '\xff':
            raise ConstructError('DotNetNullString encounted an invalid byte.')
        return None

    def _build(self, obj, stream, context, path):
        stream_write(stream, '\xff', 1)

    def _sizeof(self, context, path):
        return 1


DotNetNullString = _DotNetNullString()


class _DotNetSigToken(Adapter):
    r"""
    Adapter used to create or read a compressed token used in signatures. The token must be a typedef,
    typeref, or typespec token.

    >>> DotNetSigToken.parse('\x81\x42')
    452984912
    >>> DotNetSigToken.build(0x01000002)
    '\t'
    """
    TOKEN_ENCODE = {
        0x02: 0,
        0x01: 1,
        0x1b: 2,
    }

    def _encode(self, obj, context, path):
        encoded = self.TOKEN_ENCODE.get(obj >> 24, 3)
        if encoded is None:
            raise ConstructError('DotNetSigToken encountered a token other than typedef, typeref, or typespec')
        return ((obj & 0x00ffffff) << 2) | encoded

    def _decode(self, obj, context, path):
        if obj & 3 == 3 or obj & 0xfc00000000:
            raise ConstructError('DotNetSigToken encountered an invalid typedef, typeref, or typespec token')
        return (obj >> 2) | [0x02000000, 0x01000000, 0x1b000000][obj & 3]


DotNetSigToken = _DotNetSigToken(DotNetUInt)



"""
Description: Utility for generic, repeated functions. Expandable as needed
python version: 2.7.8
"""

import logging
from pathlib import Path
from typing import List, Optional

import pefile

logger = logging.getLogger(__name__)


def obtain_pe(file_data):
    """
    Given file data, create a pefile.PE object from the data.

    :param file_data: Input PE file data

    :return: A pefile.PE object or None
    """
    if not file_data:
        return None
    try:
        return pefile.PE(data=file_data)
    except pefile.PEFormatError as e:
        logger.debug("A pefile.PE object on the file data could not be created: {}".format(e))
        return None


def obtain_section(section_name, pe=None, file_data=None):
    """
    Obtain the section obtain for a specficied PE section of a file.

    :param section_name: The name of the section from which to extract data.
    :param pe: pefile.PE object
    :param file_data: Input file data

    :return: The PE secton object, or None.
    """
    if isinstance(section_name, str):
        section_name = section_name.encode()

    if file_data:
        pe = obtain_pe(file_data)
    if pe:
        for section in pe.sections:
            if section.Name.rstrip(b"\0") == section_name:
                return section
        return None
    else:
        return None


def obtain_section_data(section_name, pe=None, file_data=None, min_size=0) -> Optional[bytes]:
    """
    Obtain the data in a specified PE section of a file.

    :param section_name: The name of the section from which to extract data.
    :param pe: pefile.PE object
    :param file_data: Input file data
    :param min_size: The minimum acceptable size for the section_data

    :return: The PE section data, or None.
    """

    if isinstance(section_name, str):
        section_name = section_name.encode()

    if file_data:
        pe = obtain_pe(file_data)
    if pe:
        section = obtain_section(section_name, pe)
        if section:
            section_data = section.get_data()
            if len(section_data) > min_size:
                return section_data
            return None
        return None
    else:
        return None


def check_section(section_name, pe=None, file_data=None):
    """
    Check if a specified PE section exists in a file.

    :param section_name: The name of the section from which to extract data.
    :param pe: pefile.PE object
    :param file_data: Input file data

    :return: True if the section name is observed, False if it is not.
    """
    if isinstance(section_name, str):
        section_name = section_name.encode()

    if file_data:
        pe = obtain_pe(file_data)
    if pe:
        for section in pe.sections:
            if section.Name.rstrip(b"\0") == section_name:
                return True
        return False
    return False


def obtain_physical_offset(mem_offset, pe=None, file_data=None):
    """
    For an PE file, convert a provided memory offset to a raw offset.

    :param mem_offset: The memory offset to convert to a raw offset
    :param pe: pefile.PE object
    :param file_data: Input file data

    :return: Raw offset, or None.
    """
    if file_data:
        pe = obtain_pe(file_data)
    if pe:
        rva = mem_offset - pe.OPTIONAL_HEADER.ImageBase
        return pe.get_physical_by_rva(rva)
    else:
        return None


def obtain_memory_offset(raw_offset, pe=None, file_data=None):
    """
    For an PE file, convert a provided raw offset to a memory offset.

    :param raw_offset: The raw offset to convert to a memory offset
    :param pe: pefile.PE object
    :param file_data: Input file data

    :return: Memory offset, or None.
    """
    if file_data:
        pe = obtain_pe(file_data)
    if pe:
        return pe.OPTIONAL_HEADER.ImageBase + pe.get_rva_from_offset(raw_offset)
    else:
        return None


def obtain_physical_offset_x64(rel_loc, inst_end_raw, pe=None, file_data=None):
    """
    For a 64-bit PE file, pointers to data elements are relative to the end of the assembly instruction. Therefore,
    given a location (rel_loc) relative to the end of an instruction (inst_end_raw), convert the end instruction
    address to a memory offset, add that value to the relative location of the data, and convert that to a raw
    offset.

    :param rel_loc: Location of data element relative to the end of the instruction address in inst_end_raw
    :param inst_end_raw: End of an instruction address referencing the data for rel_loc
    :param pe: pefile.PE object
    :param file_data: Input file data

    :return: Raw offset for the data, or None.
    """
    if file_data:
        pe = obtain_pe(file_data)
    if pe:
        inst_end_mem = obtain_memory_offset(inst_end_raw, pe=pe)
        # Obtain the memory location of the data and convert it to a physical offset
        mem_loc = rel_loc + inst_end_mem
        return obtain_physical_offset(mem_loc, pe=pe)
    else:
        return None


def obtain_exports_list(pe=None, file_data=None) -> List[bytes]:
    """
    Obtain a list of export names for the input PE file.

    :param pe: pefile.PE object
    :param file_data: Input file data

    :return: A list of export names, or None.
    """
    if file_data:
        pe = obtain_pe(file_data)
    if pe:
        try:
            return [export.name for export in pe.DIRECTORY_ENTRY_EXPORT.symbols]
        except AttributeError:
            return []
    else:
        return []


def check_export(export_name, pe=None, file_data=None):
    """
    Check if the provided export name is in the list of exports for the file.

    :param export_name: Target export name
    :param pe: pefile.PE object
    :param file_data: Input file data

    :return bool: Indicating if provided export name is in file exports
    """
    if isinstance(export_name, str):
        export_name = export_name.encode()
    exports = obtain_exports_list(pe, file_data)
    return export_name in exports


def obtain_imported_dlls(pe=None, file_data=None) -> Optional[List[bytes]]:
    """
    Obtain a list of imported DLL names for the input PE file.

    :param pe: pefile.PE object, or None by default
    :param file_data: file data from which to create a pefile.PE object, or None by default

    :return: List of imported DLLs, or None
    """
    if file_data:
        pe = obtain_pe(file_data)
    if pe:
        try:
            return [imp.dll for imp in pe.DIRECTORY_ENTRY_IMPORT]
        except AttributeError:
            return None
    else:
        return None


def obtain_imports_list(dll_name, pe=None, file_data=None):
    """
    Obtain a list of imports from a specified DLL for the input PE file.

    :param dll_name: Name of the DLL to obtain imports from
    :param pe: pefile.PE object, or None by default
    :param file_data: file data from which to create a pefile.PE object, or None by default

    :return: List of imports from the specified DLL, or None
    """
    if file_data:
        pe = obtain_pe(file_data)
    if pe:
        try:
            for imp in pe.DIRECTORY_ENTRY_IMPORT:
                if imp.dll.lower() == dll_name.lower():
                    return [imp_func.name for imp_func in imp.imports]
        except AttributeError:
            return None
    else:
        return None


def is_imported(dll_name, func_name, pe=None, file_data=None):
    """
    Determines if a specified function is imported by the file.

    :param dll_name: Name of the DLL containing the imported function in question
    :param func_name: Name of the imported function within dll_name
    :param pe: pefile.PE object, or None by default
    :param file_data: file data from which to create a pefile.PE object, or None by default

    :return: True if function is imported
    """
    if file_data:
        pe = obtain_pe(file_data)
    if pe:
        imported_funcs = obtain_imports_list(dll_name, pe=pe)
        if imported_funcs:
            for func in imported_funcs:
                if func.lower() == func_name.lower():
                    return True
    else:
        return None


def obtain_file_ext(pe=None, file_data=None):
    """
    Attempt to return the appropriate file extension for the input PE file. Use .bin as the default if it cannot be
    recovered or None if the file is not a PE file.

    :param pe: pefile.PE object
    :param file_data: Input file data

    :return: The appropriate file extension for the PE file, .bin, or None.
    """
    if file_data:
        pe = obtain_pe(file_data)
    if pe:
        if pe.is_driver():
            return ".sys"
        elif pe.is_exe():
            return ".exe"
        elif pe.is_dll():
            return ".dll"
        else:
            return ".bin"
    else:
        return None


def is_64bit(pe=None, file_data=None):
    """
    Evaluate whether an input pefile.PE object or file data is 64-bit.

    :param pe: pefile.PE object
    :param file_data: Input file data

    :return: True if 64-bit, False if 32-bit, None if could not be determined.
    """
    if file_data:
        pe = obtain_pe(file_data)
    if pe:
        if pe.PE_TYPE == pefile.OPTIONAL_HEADER_MAGIC_PE_PLUS:
            return True
        elif pe.PE_TYPE == pefile.OPTIONAL_HEADER_MAGIC_PE:
            return False
        else:
            logger.debug("The architecture type for the file could not be determined.")
    return None


def obtain_architecture_string(pe=None, file_data=None, bitterm=True):
    """
    Obtain an architecture type string for the input PE file. Allow the bitterm variable to determine if the string
    should be in the format of "32-bit" (default) or "x86" (must specify False). Return "Undetermined" if neither.

    :param pe: pefile.PE object
    :param file_data: Input file data
    :param bitterm: Flag to determine return string type

    :return: A string representing the architecture for the input PE file.
    """
    if file_data:
        pe = obtain_pe(file_data)
    if pe:
        is64 = is_64bit(pe=pe)
        if is64:
            if bitterm:
                return "64-bit"
            else:
                return "x64"
        # Specfically check if the return value is False, because that indicates 32-bit, None indicates undetermined.
        elif is64 is False:
            if bitterm:
                return "32-bit"
            else:
                return "x86"
        else:
            return "Undetermined"
    else:
        return None


def _obtain_exif_fname(pe):
    """
    Obtain the filename from the pe.FileInfo listing of exif metadata.

    :param pe: pefile.PE object

    :return:
    """
    try:
        for file_info in pe.FileInfo:
            for entry in file_info:
                if entry.Key == b"StringFileInfo":
                    for string_table in entry.StringTable:
                        try:
                            return string_table.entries[b"OriginalFilename"]
                        except KeyError:
                            continue
    except AttributeError:
        return None


def _obtain_exportdir_fname(pe):
    """
    Obtain the filename from the export directory of the pefile.PE object.

    :param pe: pefile.PE object

    :return: The filename from the export directory, or None.
    """
    try:
        filename = pe.get_string_at_rva(pe.DIRECTORY_ENTRY_EXPORT.struct.Name)
        return filename
    except AttributeError:
        return None


def obtain_original_filename(def_stub, pe=None, file_data=None, use_arch=False, ext=".bin"):
    """
    Attempt to obtain the original filename, either from the export directory or the pe.FileInfo, of the input file.
    If the filename cannot be recovered from either of those locations, append the applicable architecture string and
    file extension to the default stub and return that value. If no pefile.PE object is provided or can be created
    from the provided file data, return the default stub appended with ".bin".

    :param def_stub: Default filename stub, sans extension, to utilize if the filename cannot be recovered.
    :param pe: pefile.PE object
    :param file_data: Input file data
    :param use_arch: Flag indicating if the file architecture should be included in the name, False by default.
    :param ext: Extension to default to if it could not be determined. (defaults to ".bin")

    :return: The recovered filename from the pe metadata or a generated filename using def_stub.
    """
    if file_data:
        pe = obtain_pe(file_data)
    if pe:
        ext = obtain_file_ext(pe=pe)
        arch = obtain_architecture_string(pe=pe, bitterm=False)
        filename = _obtain_exportdir_fname(pe) or _obtain_exif_fname(pe)
        if isinstance(filename, bytes):
            filename = filename.decode("ascii", "backslashreplace")
        if filename:
            filename = Path(filename)
            if use_arch:
                base, ext = filename.stem, filename.suffix
                filename = base + "_" + arch + ext
            return str(filename)
        else:
            return def_stub + "_" + arch + ext
    else:
        return def_stub + ext


def get_overlay_data_start_offset(pe, include_data_directories=True):
    """
    Get the offset of data appended to the file and not contained within
    the area described in the headers.

    MODIFICATIONS:
        - Use include_data_directories parameter to allow user to specify if they
          want to include the data directories in the calculation.
        - Include SECURITY table.
    """

    largest_offset_and_size = (0, 0)

    def update_if_sum_is_larger_and_within_file(offset_and_size, file_size=len(pe.__data__)):
        if sum(offset_and_size) <= file_size and sum(offset_and_size) > sum(largest_offset_and_size):
            return offset_and_size
        return largest_offset_and_size

    if hasattr(pe, "OPTIONAL_HEADER"):
        largest_offset_and_size = update_if_sum_is_larger_and_within_file(
            (pe.OPTIONAL_HEADER.get_file_offset(), pe.FILE_HEADER.SizeOfOptionalHeader)
        )

    for section in pe.sections:
        largest_offset_and_size = update_if_sum_is_larger_and_within_file(
            (section.PointerToRawData, section.SizeOfRawData)
        )

    if include_data_directories:
        for idx, directory in enumerate(pe.OPTIONAL_HEADER.DATA_DIRECTORY):
            try:
                # Security directory is special in that its VirtualAddress is actually a file offset.
                if idx == pefile.DIRECTORY_ENTRY["IMAGE_DIRECTORY_ENTRY_SECURITY"]:
                    largest_offset_and_size = update_if_sum_is_larger_and_within_file(
                        (directory.VirtualAddress, directory.Size)
                    )
                else:
                    largest_offset_and_size = update_if_sum_is_larger_and_within_file(
                        (pe.get_offset_from_rva(directory.VirtualAddress), directory.Size)
                    )
            # Ignore directories with RVA out of file
            except pefile.PEFormatError:
                continue

    if len(pe.__data__) > sum(largest_offset_and_size):
        return sum(largest_offset_and_size)

    return None


def get_overlay(pe, include_data_directories=True):
    """
    Get the data appended to the file and not contained within the area described in the headers.

    MODIFICATIONS:
        - Use include_data_directories parameter to allow user to specify if they
          want to include the data directories in the calculation.
    """

    overlay_data_offset = get_overlay_data_start_offset(pe, include_data_directories=include_data_directories)

    if overlay_data_offset is not None:
        return pe.__data__[overlay_data_offset:]

    return None


def trim(pe, include_data_directories=True):
    """
    Return the just data defined by the PE headers, removing any overlayed data.

    MODIFICATIONS:
        - Use include_data_directories parameter to allow user to specify if they
          want to include the data directories in the calculation.
    """

    overlay_data_offset = get_overlay_data_start_offset(pe, include_data_directories=include_data_directories)

    if overlay_data_offset is not None:
        return pe.__data__[:overlay_data_offset]

    return pe.__data__[:]


def is_memory_mapped(file_data):
    """
    Takes file data and tries to determine if it's a memory-mapped image (which can be squashed)
    or a normal executable file

    :param file_data: The executable file to determine whether it's a memory-mapped image

    :return True if it's a memory-mapped image, false otherwise
    """
    pe = obtain_pe(file_data)
    if pe:
        for i in range(len(pe.sections) - 1):
            if i == len(pe.sections) - 1:
                section_end = pe.OPTIONAL_HEADER.SizeOfImage
            else:
                section_end = pe.sections[i + 1].VirtualAddress
            section_start = pe.sections[i].VirtualAddress + pe.sections[i].SizeOfRawData
            if file_data[section_start:section_end] != b"\x00" * (section_end - section_start):
                return False
        return True
    return False


def squash_flat_executable(memory_mapped, pe=None):
    """
    Takes a memory mapped executable image and squashes it back down to a file that IDA can load or that can be
    executed. Note that hashes for files output by this function cannot be relied upon as valid.

    :param memory_mapped: Memory-mapped input file data
    :param pe: pefile.PE object

    :return The squashed image or None
    """
    if not pe:
        pe = obtain_pe(memory_mapped)
    if pe:
        squashed = pe.header
        for section in pe.sections:
            squashed += b"\x00" * (section.PointerToRawData - len(squashed))
            squashed += memory_mapped[section.VirtualAddress : section.VirtualAddress + section.SizeOfRawData]
        squashed += b"\x00" * (-len(squashed) & 0x1FF)
        squashed += memory_mapped[pe.OPTIONAL_HEADER.SizeOfImage :]
        return squashed
    return None


def obtain_raw_file_size(pe=None, file_data=None):
    """
    Obtain the raw file size based on the image header information. Specifically the SizeOfHeaders parameter from the
    IMAGE_OPTIONAL_HEADER, and the SizeOfRawData parameter for each PE section.

    :param pe: pefile.PE object
    :param file_data: Input file data

    :return: The raw calculated size of the file from the PE headers.
    """
    if file_data:
        pe = obtain_pe(file_data)
    if pe:
        size = pe.OPTIONAL_HEADER.SizeOfHeaders
        for section in pe.sections:
            size += section.SizeOfRawData
        return size
    return None


def has_resources(pe):
    """
    Determine if the pefile.PE object contains resources.

    :param pe: pefile.PE object to check

    :return: Boolean value on whether the pefile.PE object has a DIRECTORY_ENTRY_RESOURCE.
    """
    return hasattr(pe, "DIRECTORY_ENTRY_RESOURCE")


class Resource(object):
    """
    Object for encapsulating resource information in a simplistic format.

    :param pe: pefile.PE object
    :param entry: pe.DIRECTORY_ENTRY_RESOURCE.entries[i].directory.entries[j] resource object.
    :param dirtype: Directory name / type string
    :param data: Data for the resource
    :param offset: Starting offset to resource data.
    :param idname: ID or name string for the resource (note: always a string)
    :param rsrc_entry: String of the dirtype\idname
    """

    def __init__(self, pe, entry, dirtype):
        self._data = None
        self._pe = pe
        self._entry = entry
        self._offset = None
        self.dirtype = dirtype
        if entry.name:
            self.idname = str(entry.name)
        else:
            self.idname = str(entry.id)
        self.rsrc_entry = "%s\\%s" % (self.dirtype, self.idname)
        self.fname_stub = "%s_%s" % (self.dirtype, self.idname)

    @property
    def data(self):
        """
        Obtain the data corresponding to the resource.

        :return: Data extracted for the specified resource.
        """
        if not self._data:
            rva = self._entry.directory.entries[0].data.struct.OffsetToData
            size = self._entry.directory.entries[0].data.struct.Size
            self._data = self._pe.get_memory_mapped_image()[rva : rva + size]
            if not self._data:
                # Sometimes the get_memory_mapped_image() method failed for an unknown reason.
                # Resort to using get_data() if this happens.
                # We can't just use get_data() to start with because this method fails to pull
                # all the data for large resources.
                self._data = self._pe.get_data(rva, size)
        return self._data

    @data.setter
    def data(self, value):
        """
        Sets the data for given resource.
        """
        self._data = value

    @property
    def offset(self):
        if not self._offset:
            rva = self._entry.directory.entries[0].data.struct.OffsetToData
            self._offset = self._pe.get_physical_by_rva(rva)
        return self._offset


def iter_rsrc(pe, dirtype=None):
    """
    Iterates through resources for given pe file.

    :param pe: pefile.PE object
    :param dirtype: Optional resource directory type or name to iterate.
        defaults to iterating all directories.

    :yield: pefileutils.Resources objects
    """
    if has_resources(pe):
        for dir_entry in pe.DIRECTORY_ENTRY_RESOURCE.entries:
            # Generate dirtype
            if dir_entry.name:
                extracted_dirtype = str(dir_entry.name)
            else:
                for key, id in pefile.RESOURCE_TYPE.items():
                    if id == dir_entry.id:
                        extracted_dirtype = key
                        break
                else:
                    extracted_dirtype = str(dir_entry.id)

            # Extract entries.
            if not dirtype or str(dirtype) == extracted_dirtype:
                for entry in dir_entry.directory.entries:
                    yield Resource(pe, entry, extracted_dirtype)


def extract_all_rsrc(pe=None, file_data=None):
    """
    For a specified file, extract all resources to a list of pefileutils.Resource objects.

    :param pe: pefile.PE object
    :param file_data: Input file data

    :return: List of pefileutils.Resource objects, or an empty list.
    """
    if file_data:
        pe = obtain_pe(file_data)
    if pe:
        return list(iter_rsrc(pe))
    return []


def extract_rsrc_dir(dirtype, pe=None, file_data=None):
    """
    For a specified file, extract all resources of in a specific directory (by name or type) to a list of
    pefileutils.Resource objects.

    :param dirtype: The resource directory type or name
    :param pe: pefile.PE object
    :param file_data: Input file data

    :return: List of pefileutils.Resource objects matching the dirtype, or an empty list.
    """
    if file_data:
        pe = obtain_pe(file_data)
    if pe:
        return list(iter_rsrc(pe, dirtype=dirtype))
    return []


def extract_target_rsrc(dirtype, idname, pe=None, file_data=None):
    """
    For a specified file, extract a specific resource by name/id from a specific directory (by name or type) as a
    pefileutils.Resource object.

    :param dirtype: The resource directory type or name
    :param idname: The resource name or id, must be a string
    :param pe: pefile.PE object
    :param file_data: Input file data

    :return: A pefileutils.Resource object matching the dirtype/idname
    """
    if file_data:
        pe = obtain_pe(file_data)
    if pe:
        for rsrc in iter_rsrc(pe, dirtype=dirtype):
            assert type(rsrc.idname) == type(idname), f"{type(rsrc.idname)} != {type(idname)}"
            if rsrc.idname == idname:
                return rsrc
    return None


def check_rsrc_dir(dirtype, pe=None, file_data=None):
    """
    For a specified file, check if a specific resource directory (by name or type) exists

    :param dirtype: The resource directory type or name
    :param pe: pefile.PE object
    :param file_data: Input file data
    :return: Boolean value indicating if resource directory exists
    """

    if file_data:
        pe = obtain_pe(file_data)
    if pe:
        for _ in iter_rsrc(pe, dirtype=dirtype):
            return True
    return False



#!/usr/bin/env python
"""
technanarchy_bridge -- library to execute techanarchy ratdecoders and parse output for DC3-MWCP framework
"""
from __future__ import print_function
from future.builtins import str, zip
from six import iteritems

import os
import re
import subprocess
import sys

from io import BytesIO

# Allowing for two tabs to accommodate Publisher
TECHANARCHY_OUTPUT_RE = r"""Key: (.*?)\t{1,2} Value: (.*)"""

"""
    New key/fields added that are not current MWCP key pairs should be added to one of the lists
    below. C2 url keys are added to the domain key list.
"""
DIRECTORY_FIELD_LIST = ['Install Dir', 'InstallDir', 'InstallPath', 'Install Folder',
                        'Install Folder1', 'Install Folder2', 'Install Folder3',
                        'Folder Name', 'FolderName', 'pluginfoldername', 'nombreCarpeta']
DOMAIN_KEY_LIST = ['Domains', 'Domain', 'dns']
FILENAME_FIELD_LIST = ['InstallName', 'Install Name', 'Exe Name',
                       'Jar Name', 'JarName', 'StartUp Name', 'File Name',
                       'USB Name', 'Log File', 'Install File Name']
FILEPATH_CONCATENATE_PAIR_LIST = {'Install Path': 'Install Name',
                                  'Install Directory': 'Install File Name'}
FTP_FIELD_PAIRS = {'FTP Server': 'FTP Folder',
                   'FTPHost': 'FTPPort', 'FTPHOST': 'FTPPORT'}
INJECTIONPROCESS_FIELD_LIST = ['Process Injection', 'Injection', 'Inject Exe']
INTERVAL_FIELD_LIST = ['FTP Interval', 'Remote Delay', 'RetryInterval']
MISSIONID_FIELD_LIST = ['Campaign ID', 'CampaignID', 'Campaign Name',
                        'Campaign', 'ID', 'prefijo']
MUTEX_FIELD_LIST = ['Mutex', 'Mutex Main', 'Mutex 4', 'MUTEX',
                    'mutex', 'Mutex Grabber', 'Mutex Per']
NONC2_URL_FIELD_LIST = ['Screen Rec Link', 'WebPanel', 'Plugins']

""" The following list is used when only a password is available, that is a password without
    a corresponding username. See username below if you have a username/password pair.
"""
PASSWORD_ONLY_FIELD_LIST = ['Password', 'password']

""" Note: The username/password list are zipped together in pairs from the following
    two lists. There is a password only list above.
"""
USERNAME_FIELD_LIST = ['FTP UserName', 'FTPUserName', 'FTPUSER']
PASSWORD_FIELD_LIST = ['FTP Password', 'FTPPassword', 'FTPPASS']

REGISTRYPATH_FIELD_LIST = ['Domain', 'Reg Key', 'StartupName', 'Active X Key', 'ActiveX Key',
                           'Active X Startup', 'Registry Key', 'Startup Key', 'REG Key HKLM',
                           'REG Key HKCU', 'HKLM Value', 'RegistryKey', 'HKCUKey', 'HKCU Key',
                           'Registry Value', 'keyClase', 'regname', 'registryname',
                           'Custom Reg Key', 'Custom Reg Name', 'Custom Reg Value', 'HKCU',
                           'HKLM', 'RegKey1', 'RegKey2', 'Custom Reg Key', 'Reg Value']
VERSION_FIELD_LIST = ['Version', 'version']

"""
    End of key mapping lists
"""

def map_ta_fields(data, reporter, field_list, mwcp_key):
    for field in field_list:
        if data.get(field):
            reporter.add_metadata(mwcp_key, data[field])
def map_ta_domain_fields(data, reporter):
    for domain_key in DOMAIN_KEY_LIST:
        if domain_key in data:
            """ Hack here to handle a LuxNet case where a registry path is stored
                under the Domain key. """
            if data[domain_key].count('\\') < 2:
                domain_list = []
                if '|' in data[domain_key]:
                    """ The '|' is a separator character so strip it if
                        it is the last character so the split does not produce
                        an empty string i.e. '' """
                    domain_list = data[domain_key].rstrip('|').split('|')
                elif '*' in data[domain_key]:
                    """ The '*' is a separator character so strip it if
                        it is the last character """
                    domain_list = data[domain_key].rstrip('*').split('*')
                else:
                    domain_list = [data[domain_key]]
                for addport in domain_list:
                    if ":" in addport:
                        addr, port = addport.split(":")
                        if addr and port:
                            reporter.add_metadata(
                                "c2_socketaddress", [addr, port, "tcp"])
                    elif 'p1' in data or 'p2' in data:
                        if 'p1' in data:
                            reporter.add_metadata("c2_socketaddress", [
                                data[domain_key], data['p1'], 'tcp'])
                        if 'p2' in data:
                            reporter.add_metadata("c2_socketaddress", [
                                data[domain_key], data['p2'], 'tcp'])
                    elif 'Port' in data or 'Port1' in data or 'Port2' in data:
                        if 'Port' in data:
                            # CyberGate has a separator character in the field
                            # remove it here
                            data['Port'] = data['Port'].rstrip('|').strip('|')
                            for port in data['Port']:
                                reporter.add_metadata("c2_socketaddress", [
                                    addport, data['Port'], 'tcp'])
                        if 'Port1' in data:
                            reporter.add_metadata("c2_socketaddress", [
                                addport, data['Port1'], 'tcp'])
                        if 'Port2' in data:
                            reporter.add_metadata("c2_socketaddress", [
                                addport, data['Port2'], 'tcp'])
                    elif domain_key == 'Domain' and ("Client Control Port" in data or "Client Transfer Port" in data):
                        if "Client Control Port" in data:
                            reporter.add_metadata("c2_socketaddress", [
                                data['Domain'], data['Client Control Port'], "tcp"])
                        if "Client Transfer Port" in data:
                            reporter.add_metadata("c2_socketaddress", [data['Domain'], data[
                                'Client Transfer Port'], "tcp"])
                    else:
                        reporter.add_metadata('c2_address', data[domain_key])


def map_domainX_fields(data, reporter):
    SUFFIX_LIST = ['1', '2', '3', '4', '5', '6', '7', '8', '9', '10',
                   '11', '12', '13', '14', '15', '16', '17', '18', '19', '20']
    SPECIAL_HANDLING_LIST = ['Domain1', 'Domain2']
    for suffix in SUFFIX_LIST:
        field = 'Domain' + suffix
        if field in data:
            if data[field] != ':0':
                if ':' in data[field]:
                    address, port = data[field].split(':')
                    reporter.add_metadata('c2_socketaddress', [
                        address, port, 'tcp'])
                else:
                    if field in SPECIAL_HANDLING_LIST:
                        if "Port" in data:
                            reporter.add_metadata("c2_socketaddress", [
                                data[field], data['Port'], "tcp"])
                        elif "Port" + suffix in data:
                            # assume tcp and c2--use per scriptname
                            # customization if this doesn't hold
                            reporter.add_metadata("c2_socketaddress", [
                                data[field], data['Port' + suffix], "tcp"])
                        else:
                            reporter.add_metadata("c2_address", data[field])
                    else:
                        reporter.add_metadata('c2_address', data[field])


def map_networkgroup_nonc2_fields(data, reporter):
    map_ta_fields(data, reporter, NONC2_URL_FIELD_LIST, 'url')


def map_network_fields(data, reporter):
    map_networkgroup_nonc2_fields(data, reporter)


def map_ftp_fields(data, reporter):
    SPECIAL_HANDLING_PAIRS = {'FTP Address': 'FTP Port'}
    for host, port in iteritems(SPECIAL_HANDLING_PAIRS):
        ftpdirectory = ''
        if 'FTP Directory' in data:
            ftpdirectory = data['FTP Directory']
        mwcpkey = ''
        if host in data:
            ftpinfo = "ftp://" + data[host]
            mwcpkey = 'c2_url'
        if port in data:
            if mwcpkey:
                ftpinfo += ':' + data[port]
            else:
                ftpinfo = [data[port], 'tcp']
                mwcpkey = 'port'
        if ftpdirectory:
            if mwcpkey == 'c2_url':
                ftpinfo += '/' + ftpdirectory
                reporter.add_metadata(mwcpkey, ftpinfo)
            elif mwcpkey:
                reporter.add_metadata(mwcpkey, ftpinfo)
                reporter.add_metadata('directory', ftpdirectory)
            else:
                reporter.add_metadata('directory', ftpdirectory)
        elif mwcpkey:
            reporter.add_metadata(mwcpkey, ftpinfo)

    for address, port in iteritems(FTP_FIELD_PAIRS):
        if address in data:
            if port in data:
                reporter.add_metadata(
                    "c2_url", "ftp://" + data[address] + "/" + data[port])
            else:
                reporter.add_metadata("c2_url", "ftp://" + data[address])

def map_version_fields(data, reporter):
    map_ta_fields(data, reporter, VERSION_FIELD_LIST, 'version')


def map_mutex_fields(data, reporter):
    SPECIAL_HANDLING = 'Mutex'
    for mutex_key in MUTEX_FIELD_LIST:
        if mutex_key in data:
            if mutex_key != SPECIAL_HANDLING:
                reporter.add_metadata('mutex', data[mutex_key])
            else:
                if data[mutex_key] != 'false' and data[mutex_key] != 'true':
                    reporter.add_metadata('mutex', data[mutex_key])


def map_missionid_fields(data, reporter):
    map_ta_fields(data, reporter, MISSIONID_FIELD_LIST, 'missionid')


def map_injectionprocess_fields(data, reporter):
    map_ta_fields(data, reporter, INJECTIONPROCESS_FIELD_LIST,
                  'injectionprocess')


def map_filepath_fields(scriptname, data, reporter):
    IGNORE_SCRIPT_LIST = ['Pandora', 'Punisher']
    for pname, fname in iteritems(FILEPATH_CONCATENATE_PAIR_LIST):
        if scriptname not in IGNORE_SCRIPT_LIST:
            if pname in data:
                if fname in data:
                    reporter.add_metadata(
                        "filepath", data[pname].rstrip("\\") + "\\" + data[fname])
                else:
                    reporter.add_metadata('directory', data[pname])
            elif fname in data:
                reporter.add_metadata('filename', data[fname])
        else:
            if pname in data:
                reporter.add_metadata('directory', data[pname])
            if fname in data:
                reporter.add_metadata('filename', data[fname])


def map_directory_fields(data, reporter):
    map_ta_fields(data, reporter, DIRECTORY_FIELD_LIST, 'directory')


def map_username_password_fields(data, reporter):
    for username, password in zip(USERNAME_FIELD_LIST, PASSWORD_FIELD_LIST):
        if username in data and password in data:
            reporter.add_metadata(
                'credential', [data[username], data[password]])
        elif password in data:
            reporter.add_metadata('password', data[password])
        elif username in data:
            reporter.add_metadata('username', data[username])

    map_ta_fields(data, reporter, PASSWORD_ONLY_FIELD_LIST, 'password')


def map_interval_fields(data, reporter):
    map_ta_fields(data, reporter, INTERVAL_FIELD_LIST, 'interval')


def check_for_backslashes(ta_key, mwcp_key, data, reporter):
    IGNORE_FIELD_LIST = ['localhost', 'localhost*']
    if '\\' in data[ta_key]:
        reporter.add_metadata(mwcp_key, data[ta_key])
    elif '.' not in data[ta_key] and data[ta_key] not in IGNORE_FIELD_LIST:
        reporter.add_metadata(mwcp_key, data[ta_key])


def map_registrypath_fields(data, reporter):
    SPECIAL_HANDLING = 'Domain'
    for ta_key in REGISTRYPATH_FIELD_LIST:
        if ta_key in data:
            if ta_key == SPECIAL_HANDLING:
                check_for_backslashes(ta_key, 'registrypath', data, reporter)
            else:
                reporter.add_metadata('registrypath', data[ta_key])


def map_filename_fields(data, reporter):
    map_ta_fields(data, reporter, FILENAME_FIELD_LIST, 'filename')


def map_key_fields(data, reporter):
    if "EncryptionKey" in data:
        reporter.add_metadata("key", data["EncryptionKey"])


def map_ta_jar_fields(data, reporter):
    """This routine is for the unrecom family"""
    jarinfo = ''
    mwcpkey = ''
    if 'jarfoldername' in data:
        jarinfo = data['jarfoldername']
        mwcpkey = 'directory'
    if 'jarname' in data:
        # if a directory is added put in the \\
        if jarinfo:
            jarinfo += '\\' + data['jarname']
            mwcpkey = 'filepath'
        else:
            mwcpkey = 'filename'
            jarinfo = data['jarname']
        if 'extensionname' in data:
            jarinfo += '.' + data['extensionname']
    reporter.add_metadata(mwcpkey, jarinfo)


def map_ta_to_mwcp_keys(scriptname, data, reporter):
    """
    Updates to field mapping code belongs below here

    scriptname can be use to make per decoder customizations
    """
    map_ta_domain_fields(data, reporter)
    map_domainX_fields(data, reporter)
    map_key_fields(data, reporter)
    map_ftp_fields(data, reporter)
    map_network_fields(data, reporter)
    map_version_fields(data, reporter)
    map_mutex_fields(data, reporter)
    map_missionid_fields(data, reporter)
    map_injectionprocess_fields(data, reporter)
    map_filepath_fields(scriptname, data, reporter)
    map_directory_fields(data, reporter)
    map_username_password_fields(data, reporter)
    map_registrypath_fields(data, reporter)
    map_interval_fields(data, reporter)
    map_filename_fields(data, reporter)
    map_network_fields(data, reporter)
    map_directory_fields(data, reporter)
    """
        The following field mappings only apply to the script unrecom
    """
    if scriptname == 'unrecom':
        map_ta_jar_fields(data, reporter)


_interpreter_path = None

def interpreter_path():
    """
    Returns the path for python interpreter, assuming it can be found. Because of various
    factors (including ability to override) this may not be accurate.
    """
    global _interpreter_path
    if not _interpreter_path:
        # first try sys.executable--this is reliable most of the time but
        # doesn't work when python is embedded, ex. using wsgi mod for web
        # server
        if "python" in os.path.basename(sys.executable):
            _interpreter_path = sys.executable
        # second try sys.prefix and common executable names
        else:
            possible_path = os.path.join(sys.prefix, "python.exe")
            if os.path.exists(possible_path):
                _interpreter_path = possible_path
            possible_path = os.path.join(sys.prefix, "bin", "python")
            if os.path.exists(possible_path):
                _interpreter_path = possible_path
        # other options to consider:
        # look at some library paths, such as os.__file__, use system path to find python
        # executable that uses that library use shell and let it find python. Ex. which python
    return _interpreter_path


def run_decoder(parser, script, scriptname=""):
    """
    Run a RATdecoder and report output

    reporter: DC3-MWCP reporter object
    script: path of script to execute
    scriptname: This is the name of the decoder script, which is used for decoder specific logic.
                It defaults to the basename of the script with the .py removed

    """
    if not scriptname:
        scriptname = os.path.basename(script)[:-3]

    tempdir = parser.reporter.managed_tempdir
    outputfile = os.path.join(tempdir, "techanarchy_output")

    if interpreter_path():
        command = [interpreter_path(), script,
                   parser.file_object.file_path, outputfile]
    else:
        command = [script, parser.file_object.file_path, outputfile]

    parser.logger.info("Running %s using %s" % (scriptname, " ".join(command)))

    popen_object = subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = popen_object.communicate(None)

    termhandle = BytesIO(stdout)
    for line in termhandle:
        parser.logger.info(line.rstrip())

    termhandle = BytesIO(stderr)
    for line in termhandle:
        parser.logger.warning(line.rstrip())

    if popen_object.returncode != 0:
        parser.logger.error("Error running script. Return code: %i" %
                       popen_object.returncode)

    configlist = []
    try:
        with open(outputfile, "rb") as f:
            configlist = [line.rstrip("\n\r") for line in f]
    except Exception as e:
        parser.logger.error("Error reading script output file: %s" % str(e))

    output_re = re.compile(TECHANARCHY_OUTPUT_RE)
    output_data = {}

    for item in configlist:
        match = output_re.search(item)
        if match:
            key = match.group(1)
            value = match.group(2)
            parser.reporter.add_metadata("other", {key: value})
            if value:
                if key in output_data:
                    parser.logger.warning("collision on output key: %s" % key)
                output_data[key] = value
        else:
            parser.logger.warning("Could not parse output item: %s" % item)

    data = output_data

    map_ta_to_mwcp_keys(scriptname, data, parser.reporter)



"""
Custom Base64 related utility
"""

import base64
import logging
import sys


logger = logging.getLogger(__name__)


# Standard alphabet base on size.
_STD_ALPHA = {
    16: b'0123456789ABCDEF',
    32: b'ABCDEFGHIJKLMNOPQRSTUVWXYZ234567=',
    64: b'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=',
}


def _validate_alphabet(alphabet, type):
    """
    validate the custom alphabet
        - 64 or 65 characters
        - mappings are unique
    """
    if len(alphabet) not in (type, type+1):
        raise ValueError('invalid alphabet provided')

    if len(alphabet) != len(set(alphabet)):
        raise ValueError('mapping must be unique')

    return


def _adjust_pad(alphabet, data, decode):
    logger.warning('The padding character has not been specified in the custom alphabet')

    if not (len(data) * 8) % 6:
        logger.info('The data does not require the padding character.  continuing')
        return alphabet

    if decode:
        for char in data:
            if char not in alphabet:
                logger.info(
                    'The character "{}" does not appear in the alphabet, '
                    'but was found in the encoded data.  it will be used as the padding char'.format(char))
                return alphabet + bytes([char]) if isinstance(char, int) else char  # support for python 2 or 3
        raise ValueError('please provide a padding character to the custom alphabet')
    else:
        if b'=' not in alphabet:
            return alphabet + b'='
        else:
            raise ValueError('ERROR: please provide a padding character to the custom alphabet')


def _code(data, custom_alpha, size, decode, code_func):
    if isinstance(custom_alpha, str):
        custom_alpha = custom_alpha.encode()
    if isinstance(data, str):
        data = data.encode()
    _validate_alphabet(custom_alpha, size)
    if size != 16 and len(custom_alpha) == size:
        _adjust_pad(custom_alpha, data, decode)
    std_alpha = _STD_ALPHA[size]

    if decode:
        table = bytes.maketrans(custom_alpha, std_alpha)
        data = data.translate(table)
        return code_func(data)
    else:
        table = bytes.maketrans(std_alpha, custom_alpha)
        data = code_func(data)
        return data.translate(table)


def b64encode(data, alphabet=None):
    """
    Base64 encode
    :param data: data.
    :param alphabet: custom alphabet or standard alphabet.
    :return: base64 encoded data.

    >>> b64encode('hello world')
    'aGVsbG8gd29ybGQ='
    >>> custom_alphabet = b'EFGHQRSTUVWefghijklmnopIJKLMNOPABCDqrstuvwxyXYZabcdz0123456789+/='
    >>> b64encode('hello world', alphabet=custom_alphabet)
    'LSoXMS8BO29dMSj='
    """
    alphabet = alphabet or _STD_ALPHA[64]
    return _code(data, alphabet, 64, False, base64.b64encode)


def b64decode(data, alphabet=None):
    """
    Base64 decode (pads characters if necessary)
    :param data: base64 encoded data.
    :param alphabet: custom alphabet or standard alphabet.
    :return: base64 decoded data.

    >>> b64decode('aGVsbG8gd29ybGQ=')
    'hello world'
    >>> custom_alphabet = b'EFGHQRSTUVWefghijklmnopIJKLMNOPABCDqrstuvwxyXYZabcdz0123456789+/='
    >>> b64decode('LSoXMS8BO29dMSj=', alphabet=custom_alphabet)
    'hello world'
    >>> b64decode('LSoXMS8BO29dMSj', alphabet=custom_alphabet)
    'hello world'
    """
    alphabet = alphabet or _STD_ALPHA[64]
    # Pad the data, if necessary
    data += alphabet[len(alphabet)-1:] * ((-len(data)) % 4)
    return _code(data, alphabet, 64, True, base64.b64decode)


def b32encode(data, alphabet=None):
    """
    Base32 encodes
    :param data: data
    :param alphabet: custom alphabet or standard alphabet.
    :return: base32 encoded data.

    >>> b32encode('hello world')
    'NBSWY3DPEB3W64TMMQ======'
    >>> custom_alphabet = 'FGHIJQ345RSTUVWXYKLMABCDENOPZ267='
    >>> b32encode('hello world', alphabet=custom_alphabet)
    'VGLCEPIXJGPC6ZMUUY======'
    """
    alphabet = alphabet or _STD_ALPHA[32]
    return _code(data, alphabet, 32, False, base64.b32encode)


def b32decode(data, alphabet=None):
    """
    Base32 decode (pads characters if necessary)
    :param data: base32 encoded data.
    :param alphabet: custom alphabet or standard alphabet.
    :return: base32 decoded data.

    >>> b32decode('NBSWY3DPEB3W64TMMQ======')
    'hello world'
    >>> custom_alphabet = 'FGHIJQ345RSTUVWXYKLMABCDENOPZ267='
    >>> b32decode('VGLCEPIXJGPC6ZMUUY======', alphabet=custom_alphabet)
    'hello world'
    >>> b32decode('VGLCEPIXJGPC6ZMUUY', alphabet=custom_alphabet)
    'hello world'
    """
    alphabet = alphabet or _STD_ALPHA[32]
    # Pad the data, if necessary
    data += alphabet[len(alphabet)-1:] * ((-len(data)) % 8)
    return _code(data, alphabet, 32, True, base64.b32decode)


def b16encode(data, alphabet=None):
    """
    Base16 encodes
    :param data: data
    :param alphabet: custom alphabet or standard alphabet.
    :return: base16 encoded data.

    >>> b16encode('hello world')
    '68656C6C6F20776F726C64'
    >>> custom_alphabet = '78BDE0123F459A6C'
    >>> b16encode('hello world', alphabet=custom_alphabet)
    '131019191CB7221C2B191E'
    """
    alphabet = alphabet or _STD_ALPHA[16]
    return _code(data, alphabet, 16, False, base64.b16encode)


def b16decode(data, alphabet=None):
    """
    Base16 decode
    :param data: base16 encoded data.
    :param alphabet: custom alphabet or standard alphabet.
    :return: base16 decoded data.

    >>> b16decode('68656C6C6F20776F726C64')
    'hello world'
    >>> custom_alphabet = '78BDE0123F459A6C'
    >>> b16decode('131019191CB7221C2B191E', alphabet=custom_alphabet)
    'hello world'
    """
    alphabet = alphabet or _STD_ALPHA[16]
    return _code(data, alphabet, 16, True, base64.b16decode)



"""
Utility used for string conversions.
"""

import string
import unicodedata


def convert_to_unicode(input_value):
    if isinstance(input_value, str):
        return input_value
    elif isinstance(input_value, bytes):
        return str(input_value, encoding="latin1", errors="replace")
    else:
        return convert_to_unicode(str(input_value))


VALID_FILENAME_CHARS = "-_.() {}{}".format(string.ascii_letters, string.digits).encode("ascii")


def sanitize_filename(filename):
    """Convert given filename to sanitized version."""
    filename = convert_to_unicode(filename)
    filename = unicodedata.normalize("NFKD", filename)  # convert accented characters

    return convert_to_unicode(bytes(c for c in filename.encode("ascii", "ignore") if c in VALID_FILENAME_CHARS))



"""
Test case support for DC3-MWCP. Parser output is stored in a json file per parser. To run test cases,
parser is re-run and compared to previous results.
"""
from __future__ import print_function

# Standard imports
import json
import multiprocessing as mp
import logging
import os
import pathlib
import sys
import traceback
from timeit import default_timer

logger = logging.getLogger(__name__)

from future.builtins import str

from io import open

import mwcp
from mwcp.utils.stringutils import convert_to_unicode
from mwcp.utils import multi_proc

try:
    import pkg_resources
except ImportError:
    pkg_resources = None

# Constants
DEFAULT_EXCLUDE_FIELDS = (u"debug",)
INPUT_FILE_PATH = u"inputfilename"  # MUST BE UNICODE
FILE_EXTENSION = ".json"

# Setting encoding to utf8 is a hotfix for a larger issue
encode_params = {"encoding": "utf8", "errors": "replace"}


def multiproc_test_wrapper(args):
    """Wrapper function for running tests in multiple processes."""
    test_case = args[0]
    try:
        return test_case.run(*args[1:])
    except KeyboardInterrupt:
        return


class Tester(object):
    """DC3-MWCP Tester class"""

    def __init__(
        self, reporter, parser_names=None, nprocs=None, field_names=None, ignore_field_names=DEFAULT_EXCLUDE_FIELDS
    ):
        """

        Run tests and compare produced results to expected results.

        :param mwcp.Reporter reporter: MWCP reporter object
        :param [str] parser_names:
                A list of parser names to run tests for. If the list is empty (default),
                then test cases for all parsers will be run.
        :param [str] field_names:
                A restricted list of fields (metadata key values) that should be compared
                during testing. If the list is empty (default), then all fields, except those in
                ignore_field_names will be compared.
        :param int nprocs: Number of processes to use. (defaults to (3*num_cores)/4)
        """
        self.reporter = reporter
        self.parser_names = parser_names or [None]
        self.field_names = field_names or []
        self.ignore_field_names = ignore_field_names
        self._test_cases = None
        self._results = []  # Cached results.
        self._processed = False
        self._nprocs = nprocs or (3 * mp.cpu_count()) // 4

    def __iter__(self):
        return self._iter_results()

    def _iter_results(self):
        # First yield any cached results.
        for result in self._results:
            yield result

        # Run tests in multiprocessing pool (if not already run)
        if not self._processed:
            self._processed = True
            pool = multi_proc.TPool(processes=self._nprocs)
            test_iter = pool.imap_unordered(multiproc_test_wrapper, [(test_case,) for test_case in self.test_cases])
            pool.close()

            try:
                for result in test_iter:
                    self._results.append(result)
                    yield result
            except KeyboardInterrupt:
                pool.terminate()
                raise

    @property
    def test_cases(self):
        """Returns test cases."""
        if self._test_cases is None:
            self._test_cases = []
            for parser_name in self.parser_names:
                # We want to iterate parsers in case parser_name represents a set of parsers from different sources.
                found = False
                for source, parser in mwcp.iter_parsers(parser_name):
                    found = True
                    full_parser_name = "{}:{}".format(source.name, parser.name)
                    results_file_path = self.get_results_filepath(full_parser_name)
                    if os.path.isfile(results_file_path):
                        for expected_results in self.read_results_file(results_file_path):
                            self._test_cases.append(
                                TestCase(
                                    self.reporter,
                                    full_parser_name,
                                    expected_results,
                                    field_names=self.field_names,
                                    ignore_field_names=self.ignore_field_names,
                                )
                            )
                    else:
                        # Warn user if they are missing a test file for a parser group.
                        logger.warning("Test case file not found: {}".format(results_file_path))

                if not found and parser_name:
                    # Add a failed results if we have an orphan test.
                    self._results.append(TestResult(parser=parser_name, passed=False, errors=["Parser not found."]))
        return self._test_cases

    @property
    def total(self):
        """Returns total number of results."""
        return len(self._results) + len(self.test_cases)

    def gen_results(self, parser_name, input_file_path):
        """
        Generate JSON results for the given file using the given parser name.
        """
        self.reporter.run_parser(parser_name, input_file_path)
        self.reporter.metadata[INPUT_FILE_PATH] = convert_to_unicode(input_file_path)
        return self.reporter.metadata

    def _list_test_files(self, results_list):
        """
        Returns a list of the input file paths for the given results_list.
        """
        return [results[INPUT_FILE_PATH] for results in results_list]

    def get_results_filepath(self, name, source=None):
        """
        Returns the results file path based on the parser name provided and the
        set testcase directory.
        """
        for source, parser in mwcp.iter_parsers(name, source=source):
            file_name = parser.name + FILE_EXTENSION
            # Use hardcoded testcase directory if set.
            testcase_dir = mwcp.config.get("TESTCASE_DIR")
            if testcase_dir:
                return os.path.join(testcase_dir, file_name)

            if source.is_pkg:
                # Dynamically pull based on parser's top level module.
                test_dir = pkg_resources.resource_filename(source.path, "tests")
            else:
                # If source is a directory, assume there is a "tests" folder within it.
                test_dir = os.path.join(source.path, "tests")

            return os.path.normpath(os.path.join(test_dir, file_name))

        raise ValueError("Invalid parser: {}".format(name))

    def read_results_file(self, results_file_path):
        """
        Parses and validates the JSON results file and returns the parsed results_dict.
        """
        if not os.path.exists(results_file_path):
            results = []
        else:
            with open(results_file_path) as results_file:
                results = json.load(results_file)

        # The results file results_dict is expected to be a list of metadata dictionaries
        if not isinstance(results, list) or not all(isinstance(a, dict) for a in results):
            raise ValueError("Results file is invalid: {}".format(results_file_path))

        # Resolve input file paths.
        for testcase in results:
            # NOTE: Using PureWindowsPath to help convert a Windows path using \
            #   into / path.
            #   This helps in-case the test case was originally made with a Windows machine
            #   but is being tested on Linux.
            input_file_path = pathlib.PureWindowsPath(testcase[INPUT_FILE_PATH]).as_posix()
            # expand environment variables
            input_file_path = os.path.expandvars(input_file_path)
            # resolve variables
            input_file_path = input_file_path.format(MALWARE_REPO=mwcp.config.get("MALWARE_REPO", ""))
            # make relative paths relative to json file
            input_file_path = os.path.join(os.path.dirname(results_file_path), input_file_path)
            input_file_path = os.path.abspath(input_file_path)
            testcase[INPUT_FILE_PATH] = input_file_path

        return results

    def write_results_file(self, results_list, file_path):
        """
        Saves the JSON results list to the given file path.
        :param list[dict] results_list: JSON results list to save
        :param str file_path: Path to save the results JSON file.
        """
        # Replace references to the malware repo with a variable.
        malware_repo = mwcp.config.get("MALWARE_REPO", None)
        if malware_repo:
            for results in results_list:
                input_file_path = results[INPUT_FILE_PATH]
                if input_file_path.startswith(malware_repo):
                    input_file_path = "{MALWARE_REPO}" + input_file_path[len(malware_repo) :]
                results[INPUT_FILE_PATH] = input_file_path

        # Write updated data to results file
        # NOTE: We need to use dumps instead of dump to avoid TypeError.
        with open(file_path, "w", encoding="utf8") as results_file:
            results_file.write(str(json.dumps(results_list, indent=4, sort_keys=True)))

    def update_tests(self, force=False):
        """
        Updates existing test cases by rerunning parsers.

        :param bool force: Whether to force adding the test case even if errors are encountered
        """
        orig_level = logging.root.level
        logging.root.setLevel(logging.INFO)  # Force info level logs so test cases stay consistent.
        try:
            for parser_name in self.parser_names:
                results_file_path = self.get_results_filepath(parser_name)
                if not os.path.isfile(results_file_path):
                    logger.warning("No test case file found for parser: {}")
                    continue
                results_list = self.read_results_file(results_file_path)
                for index, file_path in enumerate(self._list_test_files(results_list)):
                    new_results = self.gen_results(parser_name, file_path)
                    if not new_results:
                        logger.warning("Empty results for {} in {}, not updating.".format(file_path, results_file_path))
                        continue
                    if self.reporter.errors and not force:
                        logger.warning("Results for {} has errors, not updating.".format(file_path))
                        continue

                    logger.info("Updating results for {} in {}".format(file_path, results_file_path))
                    results_list[index] = new_results

                self.write_results_file(results_list, results_file_path)
        finally:
            logging.root.setLevel(orig_level)

    def add_test(self, file_path, force=False, update=False):
        """
        Adds test case for given file path.

        :param str file_path: Path to input file to add.
        :param bool force: Whether to force adding the test case even if errors are encountered
        :param bool update: Whether to allow updating the test case if a test for this file already exists.
        """
        orig_level = logging.root.level
        logging.root.setLevel(logging.INFO)  # Force info level logs so test cases stay consistent.
        try:
            for parser_name in self.parser_names:
                results_file_path = self.get_results_filepath(parser_name)
                results_list = self.read_results_file(results_file_path)
                input_files = self._list_test_files(results_list)

                if file_path in input_files and not update:
                    logger.warning("Test case for {} already exists in {}".format(file_path, results_file_path))
                    continue

                new_results = self.gen_results(parser_name, file_path)
                if not new_results:
                    logger.warning("Empty results for {} in {}, not adding.".format(file_path, results_file_path))
                    continue
                if self.reporter.errors and not force:
                    logger.warning("Results for {} has errors, not adding.".format(file_path))
                    continue

                if file_path in input_files:
                    logger.info("Updating results for {} in {}".format(file_path, results_file_path))
                    index = input_files.index(file_path)
                    results_list[index] = new_results
                else:
                    logger.info("Adding results for {} in {}".format(file_path, results_file_path))
                    results_list.append(new_results)

                self.write_results_file(results_list, results_file_path)
        finally:
            logging.root.setLevel(orig_level)

    def remove_test(self, file_path):
        """Removes test case for given file path."""
        for parser_name in self.parser_names:
            results_file_path = self.get_results_filepath(parser_name)
            results_list = []
            removed = False
            for results in self.read_results_file(results_file_path):
                if results[INPUT_FILE_PATH] == file_path:
                    logger.info("Removed results for {} in {}".format(file_path, results_file_path))
                    removed = True
                else:
                    results_list.append(results)

            if removed:
                self.write_results_file(results_list, results_file_path)


class TestCase(object):
    def __init__(self, reporter, parser, expected_results, field_names=None, ignore_field_names=DEFAULT_EXCLUDE_FIELDS):
        self._reporter = reporter
        self.input_file_path = expected_results[INPUT_FILE_PATH]
        self.filename = os.path.basename(self.input_file_path)
        self.parser = parser
        self.parser_source, _, self.parser_name = parser.rpartition(":")
        self.expected_results = expected_results
        self._field_names = field_names or []
        self._ignore_field_names = ignore_field_names or []

    def run(self):
        """Run test case."""
        start_time = default_timer()

        self._reporter.run_parser(self.parser, self.input_file_path)
        self._reporter.metadata[INPUT_FILE_PATH] = convert_to_unicode(self.input_file_path)
        results = self._reporter.metadata

        comparer_results = self._compare_results(self.expected_results, results)
        passed = all(comparer.passed for comparer in comparer_results)

        done_time = default_timer()
        run_time = done_time - start_time

        return TestResult(
            parser=self.parser,
            input_file_path=self.input_file_path,
            passed=passed,
            errors=self._reporter.errors,
            debug=self._reporter.metadata.get("debug", None),
            results=comparer_results,
            run_time=run_time,
        )

    def _compare_results(self, results_a, results_b):
        """
        Compare two result sets. If the field names list is not empty,
        then only the fields (metadata key values) in the list will be compared.
        ignore_field_names fields are not compared unless included in field_names.
        """
        results = []

        # Cursory check to remove FILE_INPUT_PATH key from results since it is
        # a custom added field for test cases
        if INPUT_FILE_PATH in results_a:
            results_a = dict(results_a)
            del results_a[INPUT_FILE_PATH]
        if INPUT_FILE_PATH in results_b:
            results_b = dict(results_b)
            del results_b[INPUT_FILE_PATH]

        # Begin comparing results
        if self._field_names:
            for field_name in self._field_names:
                try:
                    comparer = self._compare_results_field(results_a, results_b, field_name)
                except:
                    comparer = ResultComparer(field_name)
                    logger.error(traceback.format_exc())
                results.append(comparer)
        else:
            for ignore_field in self._ignore_field_names:
                results_a.pop(ignore_field, None)
                results_b.pop(ignore_field, None)
            all_field_names = set(results_a.keys()).union(list(results_b.keys()))
            for field_name in all_field_names:
                try:
                    comparer = self._compare_results_field(results_a, results_b, field_name)
                except:
                    comparer = ResultComparer(field_name)
                    logger.error(traceback.format_exc())
                results.append(comparer)

        return results

    def _compare_results_field(self, results_a, results_b, field_name):
        """
        Compare the values for a single results field in the two passed in results.

        Args:
            results_a (dict): MWCP generated result for a given file using a given parser.
            results_b (dict): MWCP generated result for a given file using a given parser.
        """

        # Check if provided field_name is a valid key (based on fields.json)
        try:
            field_name_u = convert_to_unicode(field_name)
        except:
            raise Exception("Failed to convert field name '{}' to unicode.".format(field_name))

        try:
            field_type = self._reporter.fields[field_name_u]["type"]
        except:
            raise Exception("Key error. Field name '{}' was not identified as a standardized field.".format(field_name))

        # Establish value to send for comparison
        value_a = None
        value_b = None
        if field_name_u in results_a:
            value_a = results_a[field_name_u]
        if field_name_u in results_b:
            value_b = results_b[field_name_u]

        # Now compare results based on field type (see "fields.json" for more
        # details)
        if field_type == "listofstrings":
            comparer = ListOfStringsComparer(field_name_u)
            comparer.compare(value_a, value_b)
        elif field_type == "listofstringtuples":
            comparer = ListOfStringTuplesComparer(field_name_u)
            comparer.compare(value_a, value_b)
        elif field_type == "dictofstrings":
            comparer = DictOfStringsComparer(field_name_u)
            comparer.compare(value_a, value_b)
        else:
            raise Exception("Unhandled field type '{}' found for field name '{}'.".format(field_type, field_name))

        return comparer


class TestResult(object):
    def __init__(self, parser, passed, input_file_path=None, errors=None, debug=None, results=None, run_time=None):
        self.parser = parser
        self.parser_source, _, self.parser_name = parser.rpartition(":")
        self.input_file_path = input_file_path or "N/A"
        self.filename = os.path.basename(input_file_path) if input_file_path else "N/A"
        self.passed = passed
        self.errors = errors or []
        self.debug = debug or []
        self.results = results or []
        self.run_time = run_time or 0

    def print(self, failed_tests=True, passed_tests=True, json_format=False):
        """
        print test result based on provided parameters.

        :param bool failed_tests: Whether to show failed tests.
        :param bool passed_tests: Whether to show passed tests.
        :param bool json_format: Whether to format results as json.
        """
        # TODO: Do we need the json option?
        if json_format:
            passed = self.passed
            if (passed and passed_tests) or (not passed and failed_tests):
                print(json.dumps(self, indent=4, cls=MyEncoder))
        else:
            separator = ""

            filtered_output = ""
            passed = self.passed
            if passed and passed_tests:
                filtered_output += "Parser Name = {}\n".format(self.parser)
                if self.input_file_path and self.input_file_path != "N/A":
                    filtered_output += "Input Filename = {}\n".format(self.input_file_path)
                filtered_output += "Tests Passed = {}\n".format(self.passed)
            elif not passed and failed_tests:
                filtered_output += "Parser Name = {}\n".format(self.parser)
                if self.input_file_path and self.input_file_path != "N/A":
                    filtered_output += "Input Filename = {}\n".format(self.input_file_path)
                filtered_output += "Tests Passed = {}\n".format(self.passed)
                filtered_output += "Errors = {}".format("\n" if self.errors else "None\n")
                if self.errors:
                    for entry in self.errors:
                        filtered_output += "\t{0}\n".format(entry)
                filtered_output += "Debug Logs = {}".format("\n" if self.debug else "None\n")
                if self.debug:
                    for entry in self.debug:
                        filtered_output += "\t{0}\n".format(entry)
                if self.results:
                    filtered_output += "Results =\n"
                    for result in self.results:
                        if not result.passed:
                            filtered_output += "{0}\n".format(result)

            if filtered_output:
                filtered_output += "{0}\n".format(separator)
                print(filtered_output.encode("ascii", "backslashreplace").decode())


####################################################
# Comparer classes for various MWCP field types
####################################################


class ResultComparer(object):
    def __init__(self, field):
        self.field = field
        self.passed = False
        self.missing = []  # Entries found in test case but not new results
        self.unexpected = []  # Entries found in new results but not test case

    def compare(self, test_case_results=None, new_results=None):
        """Compare two result sets and document any differences."""

        self.missing = []
        self.unexpected = []

        self.field_compare(test_case_results, new_results)

        self.passed = not bool(self.missing or self.unexpected)

    def field_compare(self, test_case_results, new_results):
        """Field specific compare function."""
        # Override in child classes
        raise NotImplementedError()

    def get_report(self, json=False, tabs=1):
        """
        If json parameter is False, get report as a unicode string.
        If json parameter is True, get report as a dictionary.
        """

        if json:
            return self.__dict__
        else:
            tab = tabs * "\t"
            tab_1 = tab + "\t"
            tab_2 = tab_1 + "\t"
            report = tab + "{}:\n".format(self.field)
            report += tab_1 + "Passed: {}\n".format(self.passed)
            if self.missing:
                report += tab_1 + "Missing From New Results:\n"
                for item in self.missing:
                    report += tab_2 + "{}\n".format(convert_to_unicode(item))
            if self.unexpected:
                report += tab_1 + "Unexpected New Results:\n"
                for item in self.unexpected:
                    report += tab_2 + "{}\n".format(convert_to_unicode(item))

            return report

    def __bytes__(self):
        return self.get_report().encode("utf8")

    def __unicode__(self):
        return self.get_report()

    def __str__(self):
        if sys.version_info >= (3, 0):
            return self.__unicode__()
        else:
            return self.__bytes__()

    def __repr__(self):
        return self.__str__()


class ListOfStringsComparer(ResultComparer):
    def field_compare(self, test_case_results, new_results):
        """Compare each string in a list of strings."""
        list_test = [] if not test_case_results else test_case_results
        list_new = [] if not new_results else new_results

        self.missing += list(map(repr, set(list_test) - set(list_new)))
        self.unexpected += list(map(repr, set(list_new) - set(list_test)))


class ListOfStringTuplesComparer(ResultComparer):
    def field_compare(self, test_case_results, new_results):
        """Compare each tuple of strings in a list of tuples."""
        set_list_test = []
        set_list_new = []
        if test_case_results:
            set_list_test = [set(x) for x in test_case_results]
        if new_results:
            set_list_new = [set(x) for x in new_results]

        for set_test in set_list_test:
            if set_test not in set_list_new:
                # Append the list entry here instead of the set to preserve the
                # entries ordering
                self.missing.append(repr(test_case_results[set_list_test.index(set_test)]))

        for set_new in set_list_new:
            if set_new not in set_list_test:
                # Append the list entry here instead of the set to preserve the
                # entries ordering
                self.unexpected.append(repr(new_results[set_list_new.index(set_new)]))


class DictOfStringsComparer(ResultComparer):
    def field_compare(self, test_case_results, new_results):
        """Compare each key value pair in a dictionary of strings."""
        dict_test = {} if not test_case_results else test_case_results
        dict_new = {} if not new_results else new_results

        for key in dict_test:
            if key not in dict_new:
                self.missing.append(u"{}: {!r}".format(key, dict_test[key]))
            elif set(dict_test[key]) != set(dict_new[key]):
                self.missing.append(u"{}: {!r}".format(key, dict_test[key]))

        for key in dict_new:
            if key not in dict_test:
                self.unexpected.append(u"{}: {!r}".format(key, dict_new[key]))
            elif set(dict_new[key]) != set(dict_test[key]):
                self.unexpected.append(u"{}: {!r}".format(key, dict_new[key]))


####################################################
# JSON encoders
####################################################


class MyEncoder(json.JSONEncoder):
    def default(self, o):
        return o.__dict__



"""
Helper constructs for parsing the MIPS instruction set.
This module will be imported along with 'from mwcp.utils import construct'
and accessible from the submodule "MIPS". (e.g. construct.MIPS.lw)

reference: github.com/MIPT-ILab/mipt-mips/wiki/MIPS-Instruction-Set
"""

from .version28 import *
from .version28 import this


_REGISTERS = {
    '$zero': 0,
    '$at': 1,
    '$v0': 2, '$v1': 3,
    '$a0': 4, '$a1': 5, '$a2': 6, '$a3': 7,
    '$t0': 8, '$t1': 9, '$t2': 10, '$t3': 11, '$t4': 12, '$t5': 13, '$t6': 14, '$t7': 15,
    '$s0': 16, '$s1': 17, '$s2': 18, '$s3': 19, '$s4': 20, '$s5': 21, '$s6': 22, '$s7': 23,
    '$t8': 24, '$t9': 25,
    '$k0': 26, '$k1': 27,
    '$gp': 28, '$sp': 29, '$fp': 30, '$ra': 31,
}
_Register = Enum(BitsInteger(5), **_REGISTERS)

# I-type instruction
_I_inst = Struct(
    Embedded(BitStruct(
        'opcode' / Enum(
            BitsInteger(6),
            # NOTE: Some opcode values are reserved for other instruction formats
            # and we should let construct fail if it sees one.
            j=0x02, jal=0x03, beq=0x04, bne=0x05, blez=0x06, bgtz=0x07,
            addi=0x08, addiu=0x09, slti=0x0A, sltiu=0x0B, andi=0x0C, ori=0x0D, xori=0x0E, lui=0x0F,
            beql=0x14, bnel=0x15, blezl=0x16, bgtzl=0x17,
            daddi=0x18, daddiu=0x19, ldl=0x1A, ldr=0x1B, jalx=0x1D,
            lb=0x20, lh=0x21, lwl=0x22, lw=0x23, lbu=0x24, lhu=0x25, lwr=0x26, lwu=0x27,
            sb=0x28, sh=0x29, swl=0x2A, sw=0x2B, sdl=0x2C, sdr=0x2D, swr=0x2E, cache=0x2F,
            ll=0x30, lwc1=0x31, lwc2=0x32, pref=0x33, lld=0x34, ldc1=0x35, ldc2=0x36, ld=0x37,
            sc=0x38, swc1=0x39, swc2=0x3A, scd=0x3C, sdc1=0x3D, sdc2=0x3E, sd=0x3F,
        ),
        'src_register' / _Register,
        'target_register' / _Register,
        # 'imm_constant' / construct.BitsInteger(16)
    )),
    # Need to move immediate outside of BitStruct to create signed number.
    # (Luckly, the constant is byte aligned)
    'imm_constant' / Int16sb
)


lw = ExprValidator(_I_inst, this.opcode == 'lw')

# TODO: Create a MIPS version of ELFPointer that will account for the Global Offset Table and $gp register
# from extracted "la" psuedo instructions.



"""
    DC3-MWCP framework primary object used for execution of parsers and collection of metadata
"""
from __future__ import print_function

from future.builtins import str, open, map

import contextlib

import base64
import binascii
import hashlib
import json
import ntpath
import logging
import os
import re
import shutil
import sys
import tempfile
import codecs
import warnings

import mwcp
import mwcp.parsers
from mwcp import config
from mwcp.utils.stringutils import convert_to_unicode, sanitize_filename
from mwcp.utils import logutil

logger = logging.getLogger(__name__)
ascii_writer = codecs.getwriter("ascii")

PY3 = sys.version_info > (3,)

# pefile is now strictly optional, loaded down below so we can use
# reporter for error reporting

INFO_FIELD_ORDER = ["inputfilename", "md5", "sha1", "sha256", "compiletime"]
STANDARD_FIELD_ORDER = [
    "c2_url",
    "c2_socketaddress",
    "c2_address",
    "proxy",
    "proxy_socketaddress",
    "proxy_address",
    "url",
    "urlpath",
    "socketaddress",
    "address",
    "port",
    "listenport",
    "ftp",
    "email_address",
    "credential",
    "username",
    "password",
    "missionid",
    "useragent",
    "interval",
    "version",
    "mutex",
    "pipe",
    "guid",
    "event",
    "service",
    "servicename",
    "servicedisplayname",
    "servicedescription",
    "serviceimage",
    "servicedll",
    "injectionprocess",
    "filepath",
    "directory",
    "filename",
    "registrypath",
    "registrydata",
    "registrypathdata",
    "key",
    "rsa_private_key",
    "rsa_public_key",
    "ssl_cert_sha1",
    "base16_alphabet",
    "base32_alphabet",
    "base64_alphabet",
]


class ReporterLogHandler(logging.Handler):
    """Custom logging handler used to keep backwards compatible with legacy logging mechanism."""

    def __init__(self, reporter):
        super(ReporterLogHandler, self).__init__()
        self._reporter = reporter

    def emit(self, record):
        message = self.format(record)
        if record.levelno > logging.WARNING:
            self._reporter.errors.append(message)
        # Even though reporter uses the name "debug".. This really is an INFO level debug message.
        # (Adding true DEBUG level messages would spam our console.)
        elif logging.INFO <= record.levelno <= logging.WARNING:
            self._reporter.add_metadata("debug", message)


class Reporter(object):
    """
    Class for doing heavy lifting of parser execution and metadata reporting

    This class contains state and data about the current config parsing run, including extracted
    metadata, holding the actual sample, etc.

    Re-using an instance of this class on multiple samples is possible and should be safe, but it
    is not recommended

    Attributes:
        tempdir:
            directory where temporary files should be created. Files created in this directory should
            be deleted by parser. See managed_tempdir for mwcp managed directory
        input_file:
            the original parsed file. (instance of mwcp.FileObject)
        metadata:
            Dictionary containing the metadata extracted from the malware by the parser
        fields:
            dictionary containing the standardized fields with each field comprising an embedded
            dictionary. The 1st level keys are the field names. Under that, the keys are "description",
            "examples", and "type". See fields.json.
        errors:
            list of errors generated by framework. Generally parsers should not set these, they should
            use debug instead

    """

    URL_RE = re.compile(r"[a-z\.-]{1,40}://(?P<address>\[?[^/]+\]?)(?P<path>/[^?]+)?")
    PORT_RE = re.compile(r"[0-9]{1,5}")
    SHA1_RE = re.compile("[0-9a-fA-F]{40}")

    def __init__(
        self,
        outputdir=None,
        tempdir=None,
        disable_output_files=False,
        disable_temp_cleanup=False,
        base64_output_files=False,
        prefix_output_files=True,
    ):
        """
        Initializes the Reporter object

        :param str tempdir: sets path to temporary directory
        :param str outputdir:
            sets directory for output_file(). Should not be written to (or read from) by parsers
            directly (use tempdir)
        :param bool disable_output_files: disable writing if files to filesystem
        :param bool disable_temp_cleanup: disable cleanup (deletion) of temp files
        """

        # defaults
        self.tempdir = tempdir or tempfile.gettempdir()
        self._log_handler = None
        self.fields = {"debug": {"description": "debug", "type": "listofstrings"}}
        self.metadata = {}
        self.errors = []
        self.input_file = None

        self._managed_tempdir = None
        self._output_dir = outputdir or ""

        self._disable_output_files = disable_output_files
        self._disable_temp_cleanup = disable_temp_cleanup
        self._base64_output_files = base64_output_files
        self._prefix_output_files = prefix_output_files

        with open(config.get("FIELDS_PATH"), "rb") as f:
            self.fields = json.load(f)

    @property
    def managed_tempdir(self):
        """
        Returns the filename of a managed temporary directory. This directory will be deleted when
        parser is finished, unless tempcleanup is disabled.
        """

        if not self._managed_tempdir:
            self._managed_tempdir = tempfile.mkdtemp(dir=self.tempdir, prefix="mwcp-managed_tempdir-")

            if self._disable_temp_cleanup:
                logger.info("Using managed temp dir: {}".format(self._managed_tempdir))

        return self._managed_tempdir

    def _add_metatadata_listofstrings(self, key, value):
        if not value:
            logger.debug("no values provided for {}, skipping".format(key))
            return
        value = convert_to_unicode(value)
        obj = self.metadata.setdefault(key, [])
        if key == "debug" or value not in obj:
            obj.append(value)

        if key == "filepath":
            # use ntpath instead of os.path so we are consistent across platforms. ntpath
            # should work for both windows and unix paths. os.path works for the platform
            # you are running on, not necessarily what the malware was written for.
            # Ex. when running mwcp on linux to process windows
            # malware, os.path will fail due to not handling
            # backslashes correctly.
            self.add_metadata("filename", ntpath.basename(value))
            self.add_metadata("directory", ntpath.dirname(value))

        if key == "c2_url":
            self.add_metadata("url", value)

        if key in ("c2_address", "proxy_address"):
            self.add_metadata("address", value)

        if key == "serviceimage":
            # we use tactic of looking for first .exe in value. This is
            # not guaranteed to be reliable
            if ".exe" in value:
                self.add_metadata("filepath", value[0 : value.find(".exe") + 4])

        if key == "servicedll":
            self.add_metadata("filepath", value)

        if key == "ssl_cer_sha1":
            if not self.SHA1_RE.match(value):
                logger.error("Invalid SHA1 hash found: {!r}".format(value))

        if key in ("url", "c2_url"):
            # http://[fe80::20c:1234:5678:9abc]:80/badness
            # http://bad.com:80
            # ftp://127.0.0.1/really/bad?hostname=pwned
            match = self.URL_RE.search(value)
            if not match:
                logger.error("Error parsing as url: %s" % value)
                return

            if match.group("path"):
                self.add_metadata("urlpath", match.group("path"))

            if match.group("address"):
                address = match.group("address").rstrip(": ")
                if address.startswith("["):
                    # ipv6--something like
                    # [fe80::20c:1234:5678:9abc]:80
                    domain, found, port = address[1:].partition("]:")
                else:
                    domain, found, port = address.partition(":")

                if found:
                    if port:
                        if key == "c2_url":
                            self.add_metadata("c2_socketaddress", [domain, port, "tcp"])
                        else:
                            self.add_metadata("socketaddress", [domain, port, "tcp"])
                    else:
                        logger.error("Invalid URL {!r} found ':' at end without a port.".format(address))
                else:
                    if key == "c2_url":
                        self.add_metadata("c2_address", address)
                    else:
                        self.add_metadata("address", address)

    def _add_metadata_listofstringtuples(self, key, values):
        # Pad values that allow for shorter versions.
        expected_size = {
            "proxy": 5,
            "rsa_private_key": 8,
        }
        if key in expected_size:
            values = tuple(values) + ("",) * (expected_size[key] - len(values))

        values = list(map(convert_to_unicode, values))

        obj = self.metadata.setdefault(key, [])
        if values not in obj:
            obj.append(values)

        # Add subfield components.
        subfield_map = {
            "registrypathdata": ["registrypath", "registrydata"],
            "service": ["servicename", "servicedisplayname", "servicedescription", "serviceimage", "servicedll"],
            "credential": ["username", "password"],
        }
        if key in subfield_map:
            subfields = subfield_map[key]
            for subfield, _value in zip(subfields, values):
                if _value:
                    self.add_metadata(subfield, _value)
            if len(values) != len(subfields):
                logger.warning("Expected {} values in type {}, received {}".format(len(subfields), key, len(values)))

        # Special case validation.
        if key == "c2_socketaddress":
            self.add_metadata("socketaddress", values)
            self.add_metadata("c2_address", values[0])

        if key == "proxy_socketaddress":
            self.add_metadata("socketaddress", values)
            self.add_metadata("proxy_address", values[0])

        if key == "socketaddress":
            if len(values) != 3:
                logger.warning("Expected three values in type socketaddress, received %i" % len(values))
            self.add_metadata("address", values[0])
            self.add_metadata("port", values[1:])

        if key in ("port", "listenport"):
            if len(values) != 2:
                logger.warning("Expected two values in type %s, received %i" % (key, len(values)))
            # check for integer number and valid proto?
            match = self.PORT_RE.search(values[0])
            if match:
                portnum = int(match.group(0))
                if portnum < 0 or portnum > 65535:
                    logger.warning("Expected port to be number between 0 and 65535")
            else:
                logger.warning("Expected port to be number between 0 and 65535")
            if len(values) >= 2:
                if values[1] not in ["tcp", "udp", "icmp"]:
                    logger.warning("Expected port type to be tcp or udp (or icmp)")

        if key == "proxy":
            if len(values) != 5:
                logger.warning("Expected 5 values in type %s, received %i" % (key, len(values)))
            self.add_metadata("credential", values[:2])
            if len(values[2:]) == 1:
                self.add_metadata("proxy_address", values[2])
            else:
                self.add_metadata("proxy_socketaddress", values[2:])

        if key == "ftp":
            if len(values) != 3:
                logger.warning("Expected 3 values in type %s, received %i" % (key, len(values)))
            self.add_metadata("credential", values[:2])
            if len(values) >= 3:
                self.add_metadata("url", values[2])

        if key == "rsa_public_key":
            if len(values) != 2:
                logger.warning("Expected 3 values in type %s, received %i" % (key, len(values)))

        if key == "rsa_private_key":
            if len(values) != 8:
                logger.warning("Expected 8 values in type %s, received %i" % (key, len(values)))

    def _add_metadata_dictofstrings(self, key, value):
        # check for type of other?
        for subkey, subvalue in value.items():
            if isinstance(subvalue, (bytes, str)):
                subkey = convert_to_unicode(subkey)
                subvalue = convert_to_unicode(subvalue)
                obj = self.metadata.setdefault(key, {})
                if subkey in obj:
                    # this key already exists, we don't want to clobber so
                    # we turn into list?
                    existing_value = obj[subkey]
                    if isinstance(existing_value, list):
                        if subvalue not in obj[subkey]:
                            obj[subkey].append(subvalue)
                    elif subvalue != existing_value:
                        obj[subkey] = [existing_value, subvalue]
                else:
                    # normal insert of single value
                    obj[subkey] = subvalue
            else:
                # TODO: support inserts of lists (assuming members are strings)?
                logger.warning(
                    "Could not add object of %s to metadata under other using key %s"
                    % (str(type(subvalue[subkey])), subkey)
                )

    def add_metadata(self, key, value):
        """
        Report a metadata item

        Primary method to report metadata as a result of parsing.

        Args:
            key: string specifying the key of the metadata. Should be one of values specified in fields.json.
            value: string specifying the value of the metadata. Should be a utf-8 encoded string or a unicode object.

        """
        keyu = convert_to_unicode(key)
        if value is None or all(not _value for _value in value):
            logger.debug("no values provided for %s, skipping" % key)
            return

        if keyu not in self.fields:
            raise KeyError("Invalid field name: {}".format(keyu))

        if isinstance(value, bytes):
            value = value.decode("latin1")

        fieldtype = self.fields[keyu]["type"]

        try:
            if fieldtype == "listofstrings":
                self._add_metatadata_listofstrings(keyu, value)

            if fieldtype == "listofstringtuples":
                self._add_metadata_listofstringtuples(keyu, value)

            if fieldtype == "dictofstrings":
                self._add_metadata_dictofstrings(keyu, value)
        except Exception:
            logger.exception("Error adding metadata for key: {}".format(keyu))

    def run_parser(self, name, file_path=None, data=b"", **kwargs):
        """
        Runs specified parser on file

        :param str name: name of parser module to run (use ":" notation to specify source if necessary e.g. "mwcp-acme:Foo")
        :param str file_path: file to parse
        :param bytes data: use data as file instead of loading data from filename
        """
        self.__reset()

        # TODO: Remove all traces of the input file in the reporter!!
        #  (kept around for now because tool.py uses it for pulling file info)
        if file_path:
            with open(file_path, "rb") as f:
                self.input_file = mwcp.FileObject(
                    f.read(), self, file_name=os.path.basename(file_path), output_file=False
                )
                self.input_file.file_path = file_path
        else:
            self.input_file = mwcp.FileObject(data, self, output_file=False)

        try:
            with self.__redirect_stdout():
                found = False
                for source, parser in mwcp.iter_parsers(name):
                    found = True
                    try:
                        parser.parse(self.input_file, self)
                    except (Exception, SystemExit):
                        logger.exception(
                            "Error running parser {}:{} on {}".format(
                                source.name, parser.name, file_path or self.input_file.md5
                            )
                        )

                if not found:
                    logger.error("Could not find parsers with name: {}".format(name))
        finally:
            self.__cleanup()

    def output_file(self, data, filename, description=""):
        """
        Report a file created by the parser

        This should involve a file created by the parser and related to the malware.

        :param bytes data: The contents of the output file
        :param str filename: filename (basename) of file
        :param str description: description of the file

        :returns str: The file path of the written file if successful.
        """
        md5 = hashlib.md5(data).hexdigest()

        if self._base64_output_files:
            self.add_metadata("outputfile", [filename, description, md5, base64.b64encode(data)])
        else:
            self.add_metadata("outputfile", [filename, description, md5])

        if self._disable_output_files:
            return None

        # Create a safe filename that won't have any name collisions.
        safe_filename = sanitize_filename(filename)
        if self._prefix_output_files:
            safe_filename = f"{md5[:5]}_{safe_filename}"
        full_path = os.path.join(self._output_dir, safe_filename)

        # Make directory if it doesn't exist.
        directory = os.path.dirname(full_path)
        if directory and not os.path.isdir(directory):
            os.makedirs(directory)

        try:
            with open(full_path, "wb") as f:
                f.write(data)
            logger.debug("Output file: {}".format(full_path))
            return full_path
        except Exception as e:
            logger.error("Failed to write output file {} with error: {}".format(full_path, e))
            return None

    def format_list(self, values, key=None):

        if key == "credential" and len(values) == 2:
            return "%s:%s" % (values[0], values[1])
        elif key == "outputfile" and len(values) >= 3:
            return "%s %s" % (values[0], values[1])
        elif key in ("port", "listenport") and len(values) == 2:
            return "%s/%s" % (values[0], values[1])
        elif key == "registrykeyvalue" and len(values) == 2:
            return "%s=%s" % (values[0], values[1])
        elif key in ("socketaddress", "c2_socketaddress", "proxy_socketaddress") and len(values) == 3:
            return "%s:%s/%s" % (values[0], values[1], values[2])
        elif key == "service" and len(values) == 5:
            return ", ".join(values)
        else:
            return " ".join(values)

    def print_keyvalue(self, key, value):
        print(
            self.get_printable_key_value(key, value),
            file=ascii_writer(getattr(sys.stdout, "buffer", sys.stdout), "backslashreplace"),
        )

    def print_report(self):
        """
        Output in human readable report format
        """
        # Use sys.stdout.buffer if it exists, which is the case for Python 3 and is required
        # for writing a bytes string. Otherwise just write to whatever is at sys.stdout
        print(self.get_output_text(), file=ascii_writer(getattr(sys.stdout, "buffer", sys.stdout), "backslashreplace"))

    def get_printable_key_value(self, key, value):
        output = u""
        printkey = key

        if isinstance(value, (str, bytes)):
            output += u"{:20} {}\n".format(printkey, convert_to_unicode(value))
        else:
            for item in value:
                if isinstance(item, (str, bytes)):
                    output += u"{:20} {}\n".format(printkey, convert_to_unicode(item))
                else:
                    output += u"{:20} {}\n".format(printkey, self.format_list(item, key=key))
                printkey = u""

        return output

    def get_output_text(self):
        """
        Get data in human readable report format.
        """

        output = u""
        infoorderlist = INFO_FIELD_ORDER
        fieldorderlist = STANDARD_FIELD_ORDER

        if "inputfilename" in self.metadata:
            output += u"\n----File Information----\n\n"
            for key in infoorderlist:
                if key in self.metadata:
                    output += self.get_printable_key_value(key, self.metadata[key])

        output += u"\n----Standard Metadata----\n\n"

        for key in fieldorderlist:
            if key in self.metadata:
                output += self.get_printable_key_value(key, self.metadata[key])

        # in case we have additional fields in fields.json but the order is not
        # updated
        for key in self.metadata:
            if key not in fieldorderlist and key not in ["other", "debug", "outputfile"] and key in self.fields:
                output += self.get_printable_key_value(key, self.metadata[key])

        if "other" in self.metadata:
            output += u"\n----Other Metadata----\n\n"
            for key in sorted(list(self.metadata["other"])):
                output += self.get_printable_key_value(key, self.metadata["other"][key])

        # TODO: Should we still be showing these debug logs?
        if "debug" in self.metadata:
            output += u"\n----Debug----\n\n"
            for item in self.metadata["debug"]:
                output += u"{}\n".format(item)

        if "outputfile" in self.metadata:
            output += u"\n----Output Files----\n\n"
            for value in self.metadata["outputfile"]:
                output += self.get_printable_key_value(value[0], (value[1], value[2]))

        if self.errors:
            output += u"\n----Errors----\n\n"
            for item in self.errors:
                output += u"{}\n".format(item)

        return output

    # TODO: remove stdout redirection
    @contextlib.contextmanager
    def __redirect_stdout(self):
        """Redirects stdout temporarily to the logger."""

        class _LogWriter(object):
            def write(self, message):
                logger.info(message)

            def flush(self):
                pass

        orig_stdout = sys.stdout
        sys.stdout = _LogWriter()
        try:
            yield
        finally:
            sys.stdout = orig_stdout

    def __reset(self):
        """
        Reset all the data in the reporter object that is set during the run_parser function

        Goal is to make the reporter safe to use for multiple run_parser instances
        """
        self._managed_tempdir = None
        self.input_file = None

        self.metadata = {}
        self.errors = []

        # To keep backwards compatibility, setup log handler to add errors and debug messages to reporter.
        # TODO: Remove this when the Reporter object should no longer be responsible for logging.
        log_handler = ReporterLogHandler(self)
        logging.root.addHandler(log_handler)
        # Setup a simple format that doesn't contain any runtime variables.
        log_handler.addFilter(logutil.LevelCharFilter())
        log_handler.setFormatter(logging.Formatter("[%(level_char)s] %(message)s"))
        self._log_handler = log_handler

    def __cleanup(self):
        """
        Cleanup things
        """
        # Remove log handler.
        if self._log_handler:
            logging.root.removeHandler(self._log_handler)
            self._log_handler = None

        # Delete temporary directory.
        if not self._disable_temp_cleanup:
            if self._managed_tempdir:
                try:
                    shutil.rmtree(self._managed_tempdir, ignore_errors=True)
                except Exception as e:
                    logger.error("Failed to purge temp dir: %s, %s" % (self._managed_tempdir, str(e)))
                self._managed_tempdir = ""
        self._managed_tempdir = None

    def __del__(self):
        self.__cleanup()



"""Exposes interface for MWCP."""

import logging

# Add null handler to root logger to avoid "no handler" error when this is used as a library
logging.getLogger().addHandler(logging.NullHandler())


from mwcp.config import _config as config
from mwcp.parser import Parser
from mwcp.file_object import FileObject
from mwcp.registry import (
    register_entry_points, register_parser_directory, register_parser_package,
    iter_parsers, get_parser_descriptions, set_default_source, clear_default_source,
    ParserNotFoundError
)
from mwcp.reporter import Reporter
from mwcp.dispatcher import Dispatcher, UnableToParse, UnidentifiedFile
from mwcp.utils.logutil import setup_logging
